"""
This module is used to cache per-collection vocabulary namespace information.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2017, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

from annalist                       import layout
from annalist.exceptions            import Annalist_Error
from annalist.identifiers           import ANNAL, RDFS

from annalist.models.collectionentitycache  import (
    Cache_Error, CollectionEntityCacheObject, CollectionEntityCache
    )
from annalist.models.recordvocab            import RecordVocab

#   ---------------------------------------------------------------------------
# 
#   Collection namespace vocabulary cache class
# 
#   ---------------------------------------------------------------------------

class CollectionVocabCache(CollectionEntityCache):
    """
    This class manages and accesses namespace vocabulary cache objects for
    multiple collections.

    Per-collection cacheing is implemented by CollectionEntityCacheObject.
    """
    def __init__(self):
        """
        Initialize.

        Initializes a namespace vocabulary cache cache with no per-collection data.
        """
        super(CollectionVocabCache, self).__init__(CollectionEntityCacheObject, RecordVocab)
        return

    # Collection vocabulary namespace cache alllocation and access methods

    def set_vocab(self, coll, vocab_entity):
        """
        Save a new or updated vocabulary namespace definition
        """
        return self.set_entity(coll, vocab_entity)

    def remove_vocab(self, coll, vocab_id):
        """
        Remove vocabulary namespace from collection cache.

        Returns the vocabulary namespace entity removed if found, or None if not defined.
        """
        return self.remove_entity(coll, vocab_id)

    def get_vocab(self, coll, vocab_id):
        """
        Retrieve a vocabulary namespace description for a given Id.

        Returns a vocabulary namespace object for the specified collecion and Id.
        """
        return self.get_entity(coll, vocab_id)

    def get_vocab_from_uri(self, coll, vocab_uri):
        """
        Retrieve a vocabulary namespace description for a given URI.

        Returns a vocabulary namespace object for the specified collecion and URI.
        """
        return self.get_entity_from_uri(coll, vocab_uri)

    def get_all_vocab_ids(self, coll, altscope=None):
        """
        Returns all vocabulary namespaces currently available for a collection in the 
        indicated scope.  Default scope is vocabularies defined directly in the collection.
        """
        return self.get_all_entity_ids(coll, altscope=altscope)

    def get_all_vocabs(self, coll, altscope=None):
        """
        Returns all vocabulary namespace currently available for a collection in the 
        indicated scope.  Default scope is vocabularies defined directly in the collection.
        """
        return self.get_all_entities(coll, altscope=altscope)

# End.



"""
Annalist field group group
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import os.path
import urlparse
import shutil

import traceback
import logging
log = logging.getLogger(__name__)

from django.conf import settings

from annalist                   import layout
from annalist.exceptions        import Annalist_Error
from annalist.identifiers       import ANNAL
from annalist                   import util
from annalist.models.entity     import Entity
from annalist.models.entitydata import EntityData
from annalist.util              import extract_entity_id


class RecordGroup(EntityData):

    _entitytype     = ANNAL.CURIE.Field_group
    _entitytypeid   = layout.GROUP_TYPEID
    _entityroot     = layout.COLL_GROUP_PATH
    _entityview     = layout.COLL_GROUP_VIEW
    _entityfile     = layout.GROUP_META_FILE

    _deprecation_warning = True

    def __init__(self, parent, group_id):
        """
        Initialize a new RecordGroup object, without metadata (yet).

        parent      is the parent collection in which the group is defined.
        group_id    the local identifier for the field group
        """
        if self._deprecation_warning:
            log.warn("Instantiating _group/%s for collection %s"%(group_id, parent.get_id()))
            # log.debug("".join(traceback.format_stack()))
        super(RecordGroup, self).__init__(parent, group_id)
        self._parent = parent
        # log.debug("RecordGroup %s: dir %s"%(group_id, self._entitydir))
        return

    @classmethod
    def load(cls, parent, entityid, altscope=None, reserved_ok=False, deprecation_warning=False):
        """
        Overloaded load method with default deprecation warning
        """
        if cls._deprecation_warning:
            log.warn("Loading _group/%s for collection %s"%(entityid, parent.get_id()))
            # log.debug("".join(traceback.format_stack()))
        return super(RecordGroup, cls).load(parent, entityid, altscope=altscope, reserved_ok=reserved_ok)

    def _migrate_filenames(self):
        """
        Override EntityData method
        """
        return None

    def _migrate_values(self, entitydata):
        """
        Group description entity format migration method.

        The specification for this method is that it returns an entitydata value
        which is a copy of the supplied entitydata with format migrations applied.

        NOTE:  implementations are free to apply migrations in-place.  The resulting 
        entitydata should be exactly as the supplied data *should* appear in storage
        to conform to the current format of the data.  The migration function should 
        be idempotent; i.e.
            x._migrate_values(x._migrate_values(e)) == x._migrate_values(e)
        """
        for f in entitydata.get(ANNAL.CURIE.group_fields, []):
            field_id = extract_entity_id(f[ANNAL.CURIE.field_id])
            if field_id == "Field_render":
                f[ANNAL.CURIE.field_id] = layout.FIELD_TYPEID+"/Field_render_type"
            if field_id == "Field_type":
                f[ANNAL.CURIE.field_id] = layout.FIELD_TYPEID+"/Field_value_type"
        # Return result
        return entitydata

    def _post_update_processing(self, entitydata, post_update_flags):
        """
        Post-update processing.

        This method is called when a RecordGroup entity has been updated.  

        It invokes the containing collection method to regenerate the JSON LD context 
        for the collection to which the group belongs.
        """
        self._parent.generate_coll_jsonld_context(flags=post_update_flags)
        return entitydata

class RecordGroup_migration(RecordGroup):
    """
    Variation of RecordGroup with suppressed instantiation warning,
    used for migrating old data.
    """
    _deprecation_warning = False

    def __init__(self, parent, group_id):
        super(RecordGroup_migration, self).__init__(parent, group_id)
        return

# End.



"""
Tests for FieldDescription module
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest
from collections import OrderedDict

import logging
log = logging.getLogger(__name__)

from django.conf                        import settings
from django.test                        import TestCase # cf. https://docs.djangoproject.com/en/dev/topics/testing/tools/#assertions

from annalist                           import layout
from annalist.identifiers               import RDF, RDFS, ANNAL
from annalist.models.site               import Site
from annalist.models.collection         import Collection

from annalist.views.fielddescription    import FieldDescription, field_description_from_view_field

from annalist.views.form_utils.fieldchoice      import FieldChoice
from annalist.views.fields.render_fieldvalue    import TemplateWrapValueRenderer, ModeWrapValueRenderer
from annalist.views.fields.render_repeatgroup   import RenderRepeatGroup
from annalist.views.fields.render_placement     import (
    Placement, make_field_width, make_field_offset, make_field_display
    )

from AnnalistTestCase       import AnnalistTestCase
from tests                  import TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
from init_tests             import init_annalist_test_site, init_annalist_test_coll, resetSitedata
from entity_testentitydata  import entity_url
from entity_testsitedata    import (
    make_field_choices, no_selection,
    get_site_types, get_site_types_sorted, get_site_types_linked,
    get_site_lists, get_site_lists_sorted, get_site_lists_linked,
    get_site_views, get_site_views_sorted, get_site_views_linked,
    get_site_list_types, get_site_list_types_sorted,
    get_site_field_groups, get_site_field_groups_sorted, 
    get_site_fields, get_site_fields_sorted, 
    get_site_field_types, get_site_field_types_sorted, 
    get_site_default_entity_fields, get_site_default_entity_fields_sorted, get_site_default_entity_fields_linked, 
    )

#   -----------------------------------------------------------------------------
#
#   FieldDescription tests
#
#   -----------------------------------------------------------------------------

class FieldDescriptionTest(AnnalistTestCase):
    """
    Tests for Site object interface
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite = Site(TestBaseUri, TestBaseDir)
        self.testcoll = Collection(self.testsite, "testcoll")
        return

    def tearDown(self):
        # resetSitedata(scope="collections")
        return

    @classmethod
    def tearDownClass(cls):
        resetSitedata()
        return

    def test_FieldDescriptionTest(self):
        self.assertEqual(
            FieldDescription.__name__, "FieldDescription", 
            "Check FieldDescription class name"
            )
        return

    def test_Field_InitialValues(self):
        fd = field_description_from_view_field(
            self.testcoll, 
            { ANNAL.CURIE.field_id: layout.INITIAL_VALUES_ID }, 
            {}
            )
        expect_placement = Placement(
            width   = make_field_width(sw=12, mw=12, lw=12),
            offset  = make_field_offset(so=0, mo=0, lo=0),
            display = make_field_display(sd=True, md=True, ld=True),
            field   = "small-12 columns", 
            label   = "small-12 medium-2 columns", 
            value   = "small-12 medium-10 columns"
            )
        expect_field_desc = (
            { "field_id":                   layout.INITIAL_VALUES_ID
            , "field_name":                 layout.INITIAL_VALUES_ID
            , "field_value_type":           ""
            , "field_label":                ""
            , "field_help":                 ""
            , "field_render_type":          "Text"
            , "field_value_mode":           "Value_direct"
            , "field_property_uri":         ""
            , "field_placeholder":          ""
            , "field_default_value":        None
            , "field_placement":            expect_placement
            , "field_ref_type":             None
            , "field_choices":              None
            , "field_ref_restriction":      "ALL"
            , "field_group_ref":            None
            })
        # print repr(fd)
        self.assertDictionaryMatch(fd, expect_field_desc)
        return

    def test_Field_id(self):
        fd = field_description_from_view_field(
            self.testcoll, 
            { ANNAL.CURIE.field_id: "Field_id" }, 
            {}
            )
        expect_placement = Placement(
            width   = make_field_width(sw=12, mw=6, lw=6),
            offset  = make_field_offset(so=0, mo=0, lo=0),
            display = make_field_display(sd=True, md=True, ld=True),
            field = "small-12 medium-6 columns", 
            label = "small-12 medium-4 columns", 
            value = "small-12 medium-8 columns"
            )
        expect_field_desc = (
            { "field_id":                   "Field_id"
            , "field_name":                 "entity_id"
            , "field_value_type":           ANNAL.CURIE.EntityRef
            , "field_label":                "Field Id"
            , "field_render_type":          "EntityId"
            , "field_value_mode":           "Value_direct"
            , "field_property_uri":         ANNAL.CURIE.id
            , "field_placeholder":          "(field id)"
            , "field_default_value":        ""
            , "field_placement":            expect_placement
            , "field_ref_type":             None
            , "field_choices":              None
            , "field_ref_restriction":      "ALL"
            , "field_group_ref":            None
            })
        # print repr(fd)
        self.assertDictionaryMatch(fd, expect_field_desc)
        return

    def test_Group_field_sel(self):
        fd = field_description_from_view_field(
            self.testcoll, 
            { ANNAL.CURIE.field_id: "Group_field_sel" }, 
            {}
            )
        expect_placement = Placement(
            width   = make_field_width(sw=12, mw=6, lw=6),
            offset  = make_field_offset(so=0, mo=0, lo=0),
            display = make_field_display(sd=True, md=True, ld=True),
            field="small-12 medium-6 columns", 
            label="small-12 medium-4 columns", 
            value="small-12 medium-8 columns"
            )
        # expect_choice_id_labels = (
        #     [ ("Entity_comment",  "Comment")
        #     , ("Entity_id",       "Id"     )
        #     , ("Entity_label",    "Label"  )
        #     , ("Entity_type",     "Type"   )
        #     ])
        expect_choices = OrderedDict(
            [ (fc.id, fc) 
              for fc in no_selection("(field sel)") + 
                        get_site_default_entity_fields_linked("testcoll") 
            ])
        expect_field_desc = (
            { "field_id":                   "Group_field_sel"
            , "field_name":                 "Field_id"
            , "field_value_type":           ANNAL.CURIE.EntityRef
            , "field_label":                "Field id"
            , "field_render_type":          "Enum_optional"
            , "field_value_mode":           "Value_direct"
            , "field_property_uri":         ANNAL.CURIE.field_id
            , "field_placeholder":          "(field sel)"
            , "field_default_value":        ""
            , "field_placement":            expect_placement
            , "field_ref_type":             "_field"
            , "field_choices":              expect_choices
            , "field_ref_restriction":      "entity[annal:record_type] subtype [annal:field_entity_type]"
            , "field_group_ref":            None
            })
        # print repr(fd)
        self.assertDictionaryMatch(fd, expect_field_desc)
        return

    def test_View_fields(self):
        fd = field_description_from_view_field(
            self.testcoll, 
            { ANNAL.CURIE.field_id: "View_fields" }, 
            {}
            )
        expect_placement = Placement(
            width   = make_field_width(sw=12, mw=12, lw=12),
            offset  = make_field_offset(so=0, mo=0, lo=0),
            display = make_field_display(sd=True, md=True, ld=True),
            field = "small-12 columns", 
            label = "small-12 medium-2 columns", 
            value = "small-12 medium-10 columns"
            )
        expect_field_list = (
            [ { "annal:property_uri":       "annal:field_id"
              , "annal:field_placement":    "small:0,12;medium:0,4"
              , "annal:field_id":           "_field/View_field_sel"
              }
            , { "annal:property_uri":       "annal:property_uri"
              , "annal:field_placement":    "small:0,12;medium:4,4"
              , "annal:field_id":           "_field/View_field_property"
              }
            , { "annal:property_uri":       "annal:field_placement"
              , "annal:field_placement":    "small:0,12;medium:8,4"
              , "annal:field_id":           "_field/View_field_placement"
              }
            ])
        expect_field_desc = (
            { "field_id":               "View_fields"
            , "field_name":             "View_fields"
            , "field_value_type":       ANNAL.CURIE.View_field
            , "field_label":            "Fields"
            , "field_render_type":      "Group_Seq_Row"
            , "field_value_mode":       "Value_direct"
            , "field_property_uri":     ANNAL.CURIE.view_fields
            , "field_placeholder":      "(repeat field description)"
            , "field_default_value":    None
            , "field_placement":        expect_placement
            , "field_ref_type":         None
            , "field_choices":          None
            , "field_ref_restriction":  "ALL"
            , "field_group_ref":        None
            , "group_field_list":       expect_field_list
            })
        self.assertDictionaryMatch(fd, expect_field_desc)
        self.assertEqual(fd["field_render_type"], "Group_Seq_Row")
        # self.assertEqual(fd["field_render_colhead"], "field/annalist_head_any.html")
        # self.assertEqual(fd["field_render_colview"], "field/annalist_item_none.html")
        self.assertIsInstance(fd["field_render_colhead"], ModeWrapValueRenderer)
        self.assertIsInstance(fd["field_render_colview"], ModeWrapValueRenderer)
        self.assertIsInstance(fd["field_render_view"], ModeWrapValueRenderer)
        self.assertIsInstance(fd["field_render_edit"], ModeWrapValueRenderer)
        expect_group_details = (
            { "group_id":           "View_fields"
            , "group_label":        "Fields"
            , "group_add_label":    "Add field"
            , "group_delete_label": "Remove selected field(s)"
            })
        self.assertDictionaryMatch(fd, expect_group_details)
        self.assertEqual(len(fd["group_field_descs"]), 3)
        # Field type selector
        expect_field0_placement = Placement(
            width   = make_field_width(sw=12, mw=4, lw=4),
            offset  = make_field_offset(so=0, mo=0, lo=0),
            display = make_field_display(sd=True, md=True, ld=True),
            field="small-12 medium-4 columns", 
            label="small-12 medium-6 columns", 
            value="small-12 medium-6 columns"
            )
        expect_field0_desc = (
            { "field_id":                   "View_field_sel"
            , "field_name":                 "Field_id"
            , "field_value_type":           ANNAL.CURIE.EntityRef
            , "field_label":                "Field id"
            , "field_render_type":          "Enum_optional"
            , "field_value_mode":           "Value_direct"
            , "field_property_uri":         ANNAL.CURIE.field_id
            , "field_placement":            expect_field0_placement
            })
        self.assertDictionaryMatch(fd["group_field_descs"][0], expect_field0_desc)
        # Field property URI
        expect_field1_placement = Placement(
            width   = make_field_width(sw=12, mw=4, lw=4),
            offset  = make_field_offset(so=0, mo=4, lo=4),
            display = make_field_display(sd=True, md=True, ld=True),
            field="small-12 medium-4 columns", 
            label="small-12 medium-6 columns", 
            value="small-12 medium-6 columns"
            )
        expect_field1_desc = (
            { "field_id":                   "View_field_property"
            , "field_name":                 "Field_property"
            , "field_value_type":           ANNAL.CURIE.Identifier
            , "field_label":                "Property URI"
            , "field_render_type":          "Identifier"
            , "field_value_mode":           "Value_direct"
            , "field_property_uri":         ANNAL.CURIE.property_uri
            , "field_placement":            expect_field1_placement
            })
        self.assertDictionaryMatch(fd["group_field_descs"][1], expect_field1_desc)
        # Field placement (within group)
        expect_field2_placement = Placement(
            width   = make_field_width(sw=12, mw=4, lw=4),
            offset  = make_field_offset(so=0, mo=8, lo=8),
            display = make_field_display(sd=True, md=True, ld=True),
            field="small-12 medium-4 columns", 
            label="small-12 medium-6 columns", 
            value="small-12 medium-6 columns"
            )
        expect_field2_desc = (
            { "field_id":                   "View_field_placement"
            , "field_name":                 "Field_placement"
            , "field_value_type":           ANNAL.CURIE.Placement
            , "field_label":                "Position/size"
            , "field_render_type":          "Placement"
            , "field_value_mode":           "Value_direct"
            , "field_property_uri":         ANNAL.CURIE.field_placement
            , "field_placement":            expect_field2_placement
            })
        self.assertDictionaryMatch(fd["group_field_descs"][2], expect_field2_desc)
        return

# End.

if __name__ == "__main__":
    # Runtests in this module
    # runner = unittest.TextTestRunner(verbosity=2)
    # tests = unittest.TestSuite()
    # tests  = getSuite(select=sel)
    # if tests: runner.run(tests)
    unittest.main()



"""
This module is used to cache per-collection information about entities of some 
designated type.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2018, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

from annalist                       import layout
from annalist.exceptions            import Annalist_Error
from annalist.identifiers           import ANNAL, RDFS

#   ---------------------------------------------------------------------------
# 
#   Local helper functions
# 
#   ---------------------------------------------------------------------------


#   ---------------------------------------------------------------------------
# 
#   Error class
# 
#   ---------------------------------------------------------------------------

class Cache_Error(Annalist_Error):
    """
    Class for errors raised by closure calculations.
    """
    def __init__(self, value=None, msg="Cache_error"):
        super(Cache_Error, self).__init__(value, msg)
        return

#   ---------------------------------------------------------------------------
# 
#   Entity-cache object class
# 
#   ---------------------------------------------------------------------------

class CollectionEntityCacheObject(object):
    """
    This class is an entity cache for a specified collection and entity type.

    NOTE: entities are instantiated with respect to a specified collection,
    but the collection objects are transient (regenerated for each request), so
    the cache stores the entity values but not the instantiated entities.

    Two kinds of information are cached:

    1.  entity cache: details of all entities that are visible in
        this class, indexed by entity id and entity URI
    2.  scope cache: lists of entity ids that are visible in different scopes: used 
        when returning entity enumerations (see method "get_all_entities").

    The scope cache is populated by calls to "get_all_entities".  When a entity is 
    added to or removed from the entity cache, lacking information about the scopes 
    where it is visible, the scope cache is cleared.

    Scope values currently include "user", "all", "site"; None => "coll".
    Apart from treating None as collection local scope, the logic in this class
    treats scope names as opaque identifiers.  The scope logic is embedded mainly
    in the Entity and EntityRoot class methods "_children".
    """
    def __init__(self, coll_id, entity_cls):
        """
        Initialize a cache object for a specified collection.

        coll_id         Collection id with which the entity cache is associated.
        """
        super(CollectionEntityCacheObject, self).__init__()
        self._coll_id             = coll_id
        self._entity_cls          = entity_cls
        self._type_id             = entity_cls._entitytypeid
        self._entities_by_id      = {}
        self._entity_ids_by_uri   = {}
        self._entity_ids_by_scope = {}
        return

    def _make_entity(self, coll, entity_id, entity_values):
        """
        Internal helper method to construct an entity given its Id and values.

        coll            is collection entity to which the new identity will belong
        entity_id       is the new entity id
        entity_values   is a dictionbary containing:
                        ["parent_id"] is the id of the parent entity
                        ["data"] is a dictionary of values for the new entity

        Returns None if either Id or values evaluate as Boolean False (i.e. are None or empty).
        """
        entity = None
        if entity_id and entity_values:
            parent_id = entity_values["parent_id"]
            parent    = coll
            if coll.get_id() != parent_id:
                for parent in coll.get_alt_entities(altscope="all"):
                    if parent.get_id() == parent_id:
                        break
                else:
                    msg = (
                        "Saved parent id %s not found for entity %s/%s in collection %s"%
                        (parent_id, self._type_id, entity_id, coll.get_id())
                        )
                    log.error(msg)
                    raise ValueError(msg)
            entity = self._entity_cls._child_init(parent, entity_id)
            entity.set_values(entity_values["data"])
        return entity

    def _load_entity(self, coll, entity):
        """
        Internal helper method loads entity data to cache.

        Returns True if new entity was added.
        """
        entity_id     = entity.get_id()
        entity_uri    = entity.get_uri()
        entity_parent = entity.get_parent().get_id()
        entity_data   = entity.get_save_values()
        add_entity    = False
        if entity_id not in self._entities_by_id:
            self._entities_by_id[entity_id]     = {"parent_id": entity_parent, "data": entity_data}
            self._entity_ids_by_uri[entity_uri] = entity_id
            self._entity_ids_by_scope           = {}
            add_entity = True
        return add_entity

    def _load_entities(self, coll):
        """
        Initialize cache of entities, if not already done
        """
        if self._entities_by_id == {}:
            for entity_id in coll._children(self._entity_cls, altscope="all"):
                t = self._entity_cls.load(coll, entity_id, altscope="all")
                self._load_entity(coll, t)
        return

    def _drop_entity(self, coll, entity_id):
        """
        Drop entity from collection cache.

        Returns the entity removed, or None if not found.
        """
        entity_values = self._entities_by_id.get(entity_id, None)
        entity        = self._make_entity(coll, entity_id, entity_values)
        if entity:
            entity_uri = entity.get_uri()
            self._entities_by_id.pop(entity_id, None)
            self._entity_ids_by_uri.pop(entity_uri, None)
            self._entity_ids_by_scope = {}
        return entity

    def get_coll_id(self):
        return self._coll_id

    def set_entity(self, coll, entity):
        """
        Save a new or updated entity definition.
        """
        self._load_entities(coll)
        return self._load_entity(coll, entity)

    def remove_entity(self, coll, entity_id):
        """
        Remove entity from collection cache.

        Returns the entity removed, or None if not found.
        """
        self._load_entities(coll)
        return self._drop_entity(coll, entity_id)

    def get_entity(self, coll, entity_id):
        """
        Retrieve the entity for a given entity id.

        Returns an entity for the supplied entity Id, or None 
        if not defined in the current cache object.
        """
        self._load_entities(coll)
        entity_values = self._entities_by_id.get(entity_id, None)
        return self._make_entity(coll, entity_id, entity_values)

    def get_entity_from_uri(self, coll, entity_uri):
        """
        Retrieve an entity for a given entity URI.

        Returns an entity for the specified collecion and entuty URI,
        or None if the entity URI does not exist
        """
        self._load_entities(coll)
        entity_id = self._entity_ids_by_uri.get(entity_uri, None)
        entity    = self.get_entity(coll, entity_id)
        return entity

    def get_all_entity_ids(self, coll, altscope=None):
        """
        Returns a generator of all entity ids currently defined for a collection, 
        which may be qualified by a specified scope.

        NOTE: this method returns only those entity ids for which a record has
        been saved to the collection data storage.
        """
        self._load_entities(coll)
        scope_name = altscope or "coll"     # 'None' designates collection-local scope
        if scope_name in self._entity_ids_by_scope:
            scope_entity_ids = self._entity_ids_by_scope[scope_name]
        else:
            # Generate scope cache for named scope
            scope_entity_ids = []
            for entity_id in coll._children(self._entity_cls, altscope=altscope):
                if entity_id != layout.INITIAL_VALUES_ID:
                    scope_entity_ids.append(entity_id)
            self._entity_ids_by_scope[scope_name] = scope_entity_ids
        return scope_entity_ids

    def get_all_entities(self, coll, altscope=None):
        """
        Returns a generator of all entities currently defined for a collection, which
        may be qualified by a specified scope.

        NOTE: this method returns only those records that have actually been saved to
        the collection data storage.
        """
        scope_entity_ids = self.get_all_entity_ids(coll, altscope=altscope)
        for entity_id in scope_entity_ids:
            t = self.get_entity(coll, entity_id)
            if t:
                yield t
        return

#   ---------------------------------------------------------------------------
# 
#   Collection entity-cache class
# 
#   ---------------------------------------------------------------------------

class CollectionEntityCache(object):
    """
    This class manages multiple-collection cache objects
    """
    def __init__(self, cache_cls, entity_cls):
        """
        Initializes a value cache cache with no per-collection data.

        cache_cls   is a class object for the collaction cache objects to be used.
                    The constructor is called with collection id ent entity class
                    as parameters (see method `_get_cache`).
        entity_cls  is a class object for the type of entity to be cached.
        """
        super(CollectionEntityCache, self).__init__()
        self._cache_cls  = cache_cls
        self._entity_cls = entity_cls
        self._type_id    = entity_cls._entitytypeid
        self._caches     = {}
        return

    # Generic collection cache alllocation and access methods

    def _get_cache(self, coll):
        """
        Local helper returns a cache object for a specified collection.

        Creates a new cache object if needed.

        coll        is a collection object for which a cache object is obtained
        """
        coll_id = coll.get_id()
        if coll_id not in self._caches.keys():
            # Create and save new cache object
            self._caches[coll_id] = self._cache_cls(coll_id, self._entity_cls)
        return self._caches[coll_id]

    def flush_cache(self, coll):
        """
        Remove all cached data for a specified collection.

        Returns True if the cache object was defined, otherwise False.

        coll        is a collection object for which a cache is removed.
        """
        coll_id = coll.get_id()
        cache   = self._caches.pop(coll_id, None)
        log.info(
            "CollectionEntityCache: flushed %s cache for collection %s"%
            (self._type_id, coll_id)
            )
        return (cache is not None)

    def flush_all(self):
        """
        Remove all cached data for all collections.
        """
        self._caches = {}
        log.info(
            "CollectionEntityCache: flushed %s cache for all collections"%
            (self._type_id,)
            )
        return

    # Collection cache alllocation and access methods

    def set_entity(self, coll, entity):
        """
        Save a new or updated type definition
        """
        entity_cache = self._get_cache(coll)
        return entity_cache.set_entity(coll, entity)

    def remove_entity(self, coll, entity_id):
        """
        Remove entity from collection cache.

        Returns the entity removed if found, or None if not defined.
        """
        entity_cache = self._get_cache(coll)
        return entity_cache.remove_entity(coll, entity_id)

    def get_entity(self, coll, entity_id):
        """
        Retrieve an entity for a given entity id.

        Returns an entity object for the specified collection and entity id.
        """
        entity_cache = self._get_cache(coll)
        return entity_cache.get_entity(coll, entity_id)

    def get_entity_from_uri(self, coll, entity_uri):
        """
        Retrieve en entity for a given collection and entity URI.

        Returns an entity object for the specified collection and entity URI.
        """
        entity_cache = self._get_cache(coll)
        return entity_cache.get_entity_from_uri(coll, entity_uri)

    def get_all_entity_ids(self, coll, altscope=None):
        """
        Returns all entities currently available for a collection in the indicated scope.
        Default scope is entities defined directly in the indicated collection.
        """
        entity_cache = self._get_cache(coll)
        return entity_cache.get_all_entity_ids(coll, altscope=altscope)

    def get_all_entities(self, coll, altscope=None):
        """
        Returns all entities currently available for a collection in the indicated scope.
        Default scope is entities defined directly in the indicated collection.
        """
        entity_cache = self._get_cache(coll)
        return entity_cache.get_all_entities(coll, altscope=altscope)

# End.



"""
Entity list view
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import urlparse

import logging
log = logging.getLogger(__name__)

from django.conf                        import settings
from django.http                        import HttpResponse
from django.http                        import HttpResponseRedirect
from django.core.urlresolvers           import resolve, reverse

from annalist                           import layout
from annalist                           import message
from annalist.exceptions                import Annalist_Error
from annalist.identifiers               import RDFS, ANNAL
from annalist.util                      import (
    make_type_entity_id, split_type_entity_id, extract_entity_id,
    make_resource_url
    )

import annalist.models.entitytypeinfo as entitytypeinfo
from annalist.models.collection         import Collection
from annalist.models.recordtype         import RecordType
from annalist.models.recordtypedata     import RecordTypeData
from annalist.models.entitytypeinfo     import EntityTypeInfo, CONFIG_PERMISSIONS
from annalist.models.entityfinder       import EntityFinder

from annalist.views.uri_builder         import uri_with_params
from annalist.views.displayinfo         import DisplayInfo
from annalist.views.confirm             import ConfirmView, dict_querydict
from annalist.views.generic             import AnnalistGenericView

from annalist.views.fielddescription    import FieldDescription, field_description_from_view_field
from annalist.views.entityvaluemap      import EntityValueMap
from annalist.views.simplevaluemap      import SimpleValueMap, StableValueMap
from annalist.views.fieldlistvaluemap   import FieldListValueMap
from annalist.views.fieldvaluemap       import FieldValueMap
from annalist.views.repeatvaluesmap     import RepeatValuesMap

from annalist.views.fields.bound_field  import bound_field, get_entity_values

#   -------------------------------------------------------------------------------------------
#
#   Mapping table data (not view-specific)
#
#   -------------------------------------------------------------------------------------------

# Table used as basis, or initial values, for a dynamically generated 
# entity-value map for list displays
listentityvaluemap  = (
        [ SimpleValueMap(c='help_filename',         e=None, f=None               )
        , SimpleValueMap(c='url_type_id',           e=None, f=None               )
        , SimpleValueMap(c='url_list_id',           e=None, f=None               )
        , SimpleValueMap(c='list_choices',          e=None, f=None               )
        , SimpleValueMap(c='collection_view',       e=None, f=None               )
        , SimpleValueMap(c='default_view_id',       e=None, f=None               )
        , SimpleValueMap(c='default_view_enable',   e=None, f=None               )
        , SimpleValueMap(c='customize_view_enable', e=None, f=None               )
        , SimpleValueMap(c='search_for',            e=None, f='search_for'       )
        , SimpleValueMap(c='scope',                 e=None, f='scope'            )
        , SimpleValueMap(c='continuation_url',      e=None, f='continuation_url' )
        , SimpleValueMap(c='continuation_param',    e=None, f=None               )
        # Field data is handled separately during processing of the form description
        # Form and interaction control (hidden fields)
        ])

#   -------------------------------------------------------------------------------------------
#
#   List entities view - form rendering and POST response handling
#
#   -------------------------------------------------------------------------------------------

class EntityGenericListView(AnnalistGenericView):
    """
    View class for generic entity list view
    """

    _entityformtemplate = 'annalist_entity_list.html'

    def __init__(self):
        super(EntityGenericListView, self).__init__()
        self.help          = "entity-list-help"
        return

    # Helper functions

    def list_setup(self, coll_id, type_id, list_id, request_dict):
        """
        Assemble display information for list view request handlers
        """
        # log.info("list_setup coll_id %s, type_id %s, list_id %s"%(coll_id, type_id, list_id))
        self.collection_view_url = self.get_collection_view_url(coll_id)
        listinfo = DisplayInfo(self, "list", request_dict, self.collection_view_url)
        listinfo.get_site_info(self.get_request_host())
        listinfo.get_coll_info(coll_id)
        listinfo.get_request_type_info(type_id)
        listinfo.get_list_info(listinfo.get_list_id(listinfo.type_id, list_id))
        listinfo.check_authorization("list")
        return listinfo

    def get_list_entityvaluemap(self, listinfo, context_extra_values):
        """
        Creates an entity/value map table in the current object incorporating
        information from the form field definitions for an indicated list display.
        """
        # Locate and read view description
        entitymap  = EntityValueMap(listentityvaluemap)
        # log.debug(
        #     "EntityGenericListView.get_list_entityvaluemap entitylist %r"%
        #     listinfo.recordlist.get_values()
        #     )
        #
        # Need to generate
        # 1. 'fields':  (context receives list of field descriptions used to generate row headers)
        # 2. 'entities': (context receives a bound field that displays entry for each entity)
        #
        # NOTE - supplied entity has single field ANNAL.CURIE.entity_list (see 'get' below)
        #        entitylist template uses 'fields' from context to display headings
        list_fields = listinfo.recordlist.get(ANNAL.CURIE.list_fields, [])
        fieldlistmap = FieldListValueMap('fields', listinfo.collection, list_fields, None)
        entitymap.add_map_entry(fieldlistmap)  # For access to field headings
        repeatrows_field_descr = (
            { ANNAL.CURIE.id:                   "List_rows"
            , RDFS.CURIE.label:                 "Fields"
            , RDFS.CURIE.comment:               
                "This resource describes the repeated field description used when "+
                "displaying and/or editing a record view description"
            , ANNAL.CURIE.field_name:           "List_rows"
            , ANNAL.CURIE.field_render_type:    "RepeatListRow"
            , ANNAL.CURIE.property_uri:         ANNAL.CURIE.entity_list
            })
        repeatrows_descr = FieldDescription(
            listinfo.collection, 
            repeatrows_field_descr,
            field_list=list_fields
            )
        entitymap.add_map_entry(FieldValueMap(c="List_rows", f=repeatrows_descr))
        return entitymap

    # Helper functions assemble and return data for list of entities

    def strip_context_values(self, listinfo, entity, base_url):
        """
        Return selected values from entity data, 
        with context reference removed and entity id updated.
        """
        # entityvals = entity.get_values()
        entityvals = get_entity_values(listinfo.curr_typeinfo, entity)
        entityvals.pop('@context', None)
        entityref = make_type_entity_id(
            entityvals[ANNAL.CURIE.type_id], entityvals[ANNAL.CURIE.id]
            )
        entityvals['@id'] = base_url+entityref+"/"
        return entityvals

    def assemble_list_data(self, listinfo, scope, search_for):
        """
        Assemble and return a dict structure of JSON data used to generate
        entity list responses.
        """
        # Prepare list and entity IDs for rendering form
        selector    = listinfo.recordlist.get_values().get(ANNAL.CURIE.list_entity_selector, "")
        user_perms  = self.get_permissions(listinfo.collection)
        entity_list = (
            EntityFinder(listinfo.collection, selector=selector)
                .get_entities_sorted(
                    user_perms, type_id=listinfo.type_id, altscope=scope,
                    context={'list': listinfo.recordlist}, search=search_for
                    )
            )
        #@@
        # log.info("assemble_list_data: %r"%([e.get_id() for e in entity_list],))
        #@@
        # typeinfo = listinfo.curr_typeinfo
        base_url = self.get_collection_base_url(listinfo.coll_id)
        list_url = self.get_list_url(
            listinfo.coll_id, listinfo.list_id,
            type_id=listinfo.type_id,
            scope=scope,
            search=search_for
            )
        entityvallist = [ self.strip_context_values(listinfo, e, base_url) for e in entity_list ]
        # print "@@@@ type_id %s"%(listinfo.type_id,)
        # print "@@@@ entity_list %r"%(entity_list,)
        # print "@@@@ entityvallist %r"%(entityvallist,)
        # log.debug("@@ listinfo.list_id %s, coll base_url %s"%(listinfo.list_id, base_url))
        log.info(
            "EntityListDataView.assemble_list_data: list_url %s, base_url %s, context_url %s"%
            (list_url, base_url, base_url+layout.COLL_CONTEXT_FILE)
            )
        jsondata = (
            { '@id':            list_url
            , '@context': [
                { "@base":  base_url },
                base_url+layout.COLL_CONTEXT_FILE
                ]
            , ANNAL.CURIE.entity_list:  entityvallist
            })
        # print "@@@@ assemble_list_data: jsondata %r"%(jsondata,)
        return jsondata

    # GET

    def get(self, request, coll_id=None, type_id=None, list_id=None):
        """
        Create a form for listing entities.
        """
        scope      = request.GET.get('scope',  None)
        search_for = request.GET.get('search', "")
        log.info(
            "views.entitylist.get:  coll_id %s, type_id %s, list_id %s, scope %s, search '%s'"%
            (coll_id, type_id, list_id, scope, search_for)
            )
        log.log(settings.TRACE_FIELD_VALUE, "  %s"%(self.get_request_path()))
        listinfo    = self.list_setup(coll_id, type_id, list_id, request.GET.dict())
        if listinfo.http_response:
            return listinfo.http_response
        self.help_markdown = listinfo.recordlist.get(RDFS.CURIE.comment, None)
        log.debug("listinfo.list_id %s"%listinfo.list_id)
        # Prepare list and entity IDs for rendering form
        try:
            entityvallist = self.assemble_list_data(listinfo, scope, search_for)
            # Set up initial view context
            context_extra_values = (
                { 'continuation_url':       listinfo.get_continuation_url() or ""
                , 'continuation_param':     listinfo.get_continuation_param()
                , 'request_url':            self.get_request_path()
                , 'scope':                  scope
                , 'url_type_id':            type_id
                , 'url_list_id':            list_id
                , 'search_for':             search_for
                , 'list_choices':           self.get_list_choices_field(listinfo)
                , 'collection_view':        self.collection_view_url
                , 'default_view_id':        listinfo.recordlist[ANNAL.CURIE.default_view]
                , 'default_view_enable':    'disabled="disabled"'
                , 'customize_view_enable':  'disabled="disabled"'
                , 'collection':             listinfo.collection
                })
            if listinfo.authorizations['auth_config']:
                context_extra_values['customize_view_enable'] = ""
                if list_id:
                    context_extra_values['default_view_enable']   = ""
            entityvaluemap = self.get_list_entityvaluemap(listinfo, context_extra_values)
            listcontext = entityvaluemap.map_value_to_context(
                entityvallist,
                **context_extra_values
                )
            listcontext.update(listinfo.context_data())
        except Exception as e:
            log.exception(str(e))
            return self.error(
                dict(self.error500values(),
                    message=str(e)+" - see server log for details"
                    )
                )
        # log.info("EntityGenericListView.get listcontext %r"%(listcontext))
        # Generate and return form data
        json_redirect_url   = make_resource_url("", self.get_request_path(), layout.ENTITY_LIST_FILE)
        turtle_redirect_url = make_resource_url("", self.get_request_path(), layout.ENTITY_LIST_TURTLE)
        return (
            self.render_html(listcontext, self._entityformtemplate)
            or 
            self.redirect_json(json_redirect_url)
            or
            self.redirect_turtle(turtle_redirect_url)
            or
            self.error(self.error406values())
            )

    # POST

    def post(self, request, coll_id=None, type_id=None, list_id=None):
        """
        Handle response from dynamically generated list display form.
        """
        log.info("views.entitylist.post: coll_id %s, type_id %s, list_id %s"%(coll_id, type_id, list_id))
        log.log(settings.TRACE_FIELD_VALUE, "  %s"%(self.get_request_path()))
        # log.log(settings.TRACE_FIELD_VALUE, "  form data %r"%(request.POST))
        listinfo = self.list_setup(coll_id, type_id, list_id, request.POST.dict())
        if listinfo.http_response:
            return listinfo.http_response
        if 'close' in request.POST:
            return HttpResponseRedirect(listinfo.get_continuation_url() or self.collection_view_url)

        # Process requested action
        action          = None
        redirect_path   = None
        redirect_cont   = listinfo.get_continuation_here()
        redirect_params = {}
        entity_ids      = request.POST.getlist('entity_select')
        log.debug("entity_ids %r"%(entity_ids))
        if len(entity_ids) > 1:
            listinfo.display_error_response(message.TOO_MANY_ENTITIES_SEL)
        else:
            entity_type = type_id or listinfo.get_list_type_id()
            entity_id   = None
            if len(entity_ids) == 1:
                (entity_type, entity_id) = split_type_entity_id(entity_ids[0], entity_type)
            if "new" in request.POST:
                action        = "new"
                redirect_path = listinfo.get_new_view_uri(coll_id, entity_type)
            if "copy" in request.POST:
                action = "copy"
                if not entity_id:
                    listinfo.display_error_response(message.NO_ENTITY_FOR_COPY)
                else:
                    redirect_path = listinfo.get_edit_view_uri(
                        coll_id, entity_type, entity_id, action
                        )
            if "edit" in request.POST:
                action = "edit"
                if not entity_id:
                    listinfo.display_error_response(message.NO_ENTITY_FOR_EDIT)
                else:
                    redirect_path = listinfo.get_edit_view_uri(
                        coll_id, entity_type, entity_id, action
                        )
            if "delete" in request.POST:
                action = "delete"
                confirmed_deletion_uri = self.view_uri(
                    "AnnalistEntityDataDeleteView", 
                    coll_id=coll_id, type_id=entity_type
                    )
                return listinfo.confirm_delete_entity_response(
                    entity_type, entity_id, 
                    confirmed_deletion_uri
                    )
            if "default_view" in request.POST:
                #@@
                # auth_check = self.form_action_auth("config", listinfo.collection, CONFIG_PERMISSIONS)
                #@@
                auth_check = listinfo.check_authorization("config")
                if auth_check:
                    return auth_check
                listinfo.collection.set_default_list(list_id)
                listinfo.add_info_message(
                    message.DEFAULT_LIST_UPDATED%{'coll_id': coll_id, 'list_id': list_id}         
                    )
                redirect_path, redirect_params = listinfo.redisplay_path_params()
                redirect_cont   = listinfo.get_continuation_next()
            if ( ("list_type" in request.POST) or ("list_all"  in request.POST) ):
                action          = "list"
                redirect_path   = self.get_list_url(
                    coll_id, extract_entity_id(request.POST['list_choice']),
                    type_id=None if "list_all" in request.POST else type_id
                    )
                redirect_params = dict(
                    scope="all" if "list_scope_all" in request.POST else None,
                    search=request.POST['search_for']
                    )
                redirect_cont   = listinfo.get_continuation_next()
                # redirect_cont   = None
            if "customize" in request.POST:
                action        = "config"
                redirect_path = self.view_uri(
                            "AnnalistCollectionEditView", 
                            coll_id=coll_id
                            )
        if redirect_path:
            if redirect_cont:
                redirect_params.update(
                    { "continuation_url": redirect_cont }
                    )
            listinfo.redirect_response(
                redirect_path, redirect_params=redirect_params, action=action
                )
            # return (
            #     listinfo.check_authorization(action) or
            #     HttpResponseRedirect(redirect_uri)
            #     )
        if listinfo.http_response:
            return listinfo.http_response


        # Report unexpected form data
        # This shouldn't happen, but just in case...
        # Redirect to continuation with error
        log.error("Unexpected form data posted to %s: %r"%(request.get_full_path(), request.POST))
        err_values = self.error_params(
            message.UNEXPECTED_FORM_DATA%(request.POST), 
            message.SYSTEM_ERROR
            )
        redirect_uri = uri_with_params(listinfo.get_continuation_next(), err_values)
        return HttpResponseRedirect(redirect_uri)

    def get_list_choices_field(self, listinfo):
        """
        Returns a bound_field object that displays as a list-choice selection drop-down.
        """
        # @@TODO: Possibly create FieldValueMap and return map_entity_to_context value? 
        #         or extract this logic and share?  See also entityedit view choices
        field_description = field_description_from_view_field(
            listinfo.collection, 
            { ANNAL.CURIE.field_id: "List_choice"
            , ANNAL.CURIE.field_placement: "small:0,12;medium:0,9" },
            {}
            )
        entityvals = { field_description['field_property_uri']: listinfo.list_id }
        return bound_field(field_description, entityvals)

    def get_list_url(self, coll_id, list_id, type_id=None, scope=None, search=None, query_params={}):
        """
        Return a URL for accessing the current list display
        """
        list_uri_params = (
            { 'coll_id': coll_id
            , 'list_id': list_id
            })
        if type_id:
            list_uri_params['type_id']  = type_id
        if scope:
            query_params = dict(query_params, scope=scope)
        if search:
            query_params = dict(query_params, search=search)
        list_url = (
            uri_with_params(
                self.view_uri("AnnalistEntityGenericList", **list_uri_params),
                query_params
                )
            )
        return list_url

# End.



"""
Tests for site data.

This module contains tests that are intended to test the validity, 
completeness and consistency of Annalist-defined site-wide data.
It assumes that other test modules confirm operation of the 
underlying web site server code.

The general pattern of these tests is to render a web page and then to 
Check that the rendered page contains appropriate contents as defined 
by the Annalist site data.

A key reason for this module's existence is to catch any regressions in the
presentation of views that have been manually checked for correctness that 
may be caused by changes to the underlying site data structures.
It is not intended to check details of web page presentation.

(The first couple of tests also check aspects of the test site data setup.)
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest

import logging
log = logging.getLogger(__name__)

from django.conf                import settings
from django.contrib.auth.models import User
from django.test                import TestCase
from django.test.client         import Client

from bs4                        import BeautifulSoup

from annalist                               import layout
from annalist.identifiers                   import RDF, RDFS, ANNAL
from annalist.util                          import extract_entity_id

from annalist.models.site                   import Site
from annalist.models.collection             import Collection
from annalist.models.recordtype             import RecordType
from annalist.models.recordlist             import RecordList
from annalist.models.recordview             import RecordView
from annalist.models.recordfield            import RecordField

from annalist.views.fields.find_renderers   import is_repeat_field_render_type
from annalist.views.fields.render_placement import get_placement_options
from annalist.views.form_utils.fieldchoice  import FieldChoice, get_choice_labels

from AnnalistTestCase       import AnnalistTestCase
from tests                  import TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
from init_tests             import init_annalist_test_site, init_annalist_test_coll, resetSitedata
from entity_testutils       import (
    site_view_url,
    collection_view_url,
    collection_edit_url,
    collection_entity_list_url,
    collection_entity_view_url,
    collection_entity_edit_url,
    create_user_permissions, create_test_user
    )
from entity_testcolldata    import (
    collectiondata_url, collectiondata_resource_url,
    collectiondata_view_url, collectiondata_view_resource_url,
    collectiondata_value_keys, collectiondata_load_keys,
    collectiondata_create_values, collectiondata_values, collectiondata_read_values,
    collectiondata_view_form_data
    )
from entity_testsitedata    import (
    make_field_choices, no_selection,
    get_site_types,        get_site_types_sorted,
    get_site_lists,        get_site_lists_sorted,
    get_site_views,        get_site_views_sorted,
    get_site_fields,       get_site_fields_sorted, 
    get_site_field_groups, get_site_field_groups_sorted, 
    get_site_list_types,   get_site_list_types_sorted, 
    get_site_field_types,  get_site_field_types_sorted,
    # Fields by entity type
    get_site_type_fields,  get_site_type_fields_sorted, 
    get_site_list_fields,  get_site_list_fields_sorted, 
    get_site_view_fields,  get_site_view_fields_sorted, 
    get_site_vocab_fields, get_site_vocab_fields_sorted, 
    get_site_field_fields, get_site_field_fields_sorted,
    get_site_group_fields, get_site_group_fields_sorted,
    get_site_user_fields,  get_site_user_fields_sorted, 
    )

#   -----------------------------------------------------------------------------
#
#   Site data tests
#
#   -----------------------------------------------------------------------------

class AnnalistSiteDataTest(AnnalistTestCase):
    """
    Tests for Annalist site data completeness and consistency
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite  = Site(TestBaseUri, TestBaseDir)
        self.coll1     = Collection.load(self.testsite, "coll1")
        self.local_types = make_field_choices(
            [ ("_type/type1", "RecordType coll1/type1")
            , ("_type/type2", "RecordType coll1/type2")
            ])
        self.local_lists = make_field_choices(
            [ ("_list/list1", "RecordList coll1/list1")
            , ("_list/list2", "RecordList coll1/list2")
            ])
        self.local_views = make_field_choices(
            [ ("_view/view1", "RecordView coll1/view1")
            , ("_view/view2", "RecordView coll1/view2")
            ])
        self.types_expected = get_site_types_sorted() + self.local_types
        self.lists_expected = get_site_lists_sorted() + self.local_lists
        self.views_expected = get_site_views_sorted() + self.local_views
        self.list_types_expected   = get_site_list_types_sorted()
        self.render_types_expected = get_site_field_types_sorted()
        self.grouprefs_expected    = get_site_field_groups_sorted()
        self.placements_expected   = get_placement_options()

        # Login with admin permissions
        permissions=["VIEW", "CREATE", "UPDATE", "DELETE", "CONFIG", "ADMIN"]
        create_test_user(
            self.coll1, 
            "testuser", "testpass", user_permissions=permissions
            )
        self.client = Client(HTTP_HOST=TestHost)
        loggedin = self.client.login(username="testuser", password="testpass")
        self.assertTrue(loggedin)
        return

    def tearDown(self):
        # resetSitedata()
        return

    @classmethod
    def tearDownClass(cls):
        resetSitedata()
        return

    # --------------------------------------------------------------------------
    # Helper functions
    # --------------------------------------------------------------------------

    # Dereference URI and return BeautifulSoup object for the returned HTML
    def get_page(self, uri):
        r = self.client.get(uri)
        if r.status_code == 302:
            r = self.client.get(r['location'])
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        s = BeautifulSoup(r.content, "html.parser")
        return s

    # Test named input field has specified type and value
    def check_input_type_value(self, 
            s, name, 
            input_type, input_value=None, input_text=None):
        tag_name = "input"
        tag_type = input_type
        if input_type == "textarea":
            tag_name = "textarea"
            tag_type = None
        input_elem = s.form.find(attrs={"name": name})
        self.assertEqual(input_elem.name, tag_name)
        if tag_type:
            self.assertEqual(input_elem['type'], tag_type)
        if input_value is not None:
            self.assertEqual(input_elem['value'], input_value)
        if input_text is not None:
            self.assertEqual(input_elem.string, input_text)
        return

    # Convert selected Unicode characters to HTMLentity form (undoing changes wrought by BeautifulSoup)
    @staticmethod
    def html_encode(s):
        # See: http://stackoverflow.com/a/1919221/324122
        entity_map = (
            { u'\u2588': u"&block;"
            , u'\u2591': u"&blk14;"
            })
        for u, e in entity_map.iteritems():
            s = s.replace(u, e)
        return s

    # Test named input is <select> with specified options and selected value
    # self.check_select_field(s, "typelist", self.local_types, None)
    def check_select_field(self, s, name, options, selection):
        select_elem  = s.find("select", attrs={"name": name})
        option_values_here     = (
            [self.html_encode(o.get('value',"@@novalue:%s"%o.string))
             for o in select_elem.find_all("option")
            ])
        option_values_expected = [ fc.value for fc in options ]
        if option_values_here != option_values_expected:
            log.info("option_values_here:     %r"%(option_values_here,))
            log.info("option_values_expected: %r"%(option_values_expected,))
        self.assertEqual(option_values_here, option_values_expected)
        option_labels_here     = (
            [self.html_encode(o.string) for o in select_elem.find_all("option")]
            )
        option_labels_expected = get_choice_labels(options)
        if option_labels_here != option_labels_expected:
            log.info("option_labels_here:     %r"%(option_labels_here,))
            log.info("option_labels_expected: %r"%(option_labels_expected,))
        self.assertEqual(option_labels_here, option_labels_expected)
        if selection is not None:
            find_selection = select_elem.find("option", selected=True)
            if find_selection is None:
                self.fail("No selection %s for %s"%(selection, select_elem))
            self.assertEqual(find_selection['value'], selection)
        return

    def check_row_column(self, row_data, colnum, row_expected):
        if row_expected[colnum] is not None:
            e = row_expected[colnum]
            i = (row_data
                    .find("div", class_="row")
                    .find_all("div", class_="columns")[colnum]
                    .stripped_strings
                )
            f = next(i, "(@@no value@@)")
            if f != e:
                print "@@ Expected: %r, found %r"%(e, f)
                cols = (row_data
                        .find("div", class_="row")
                        .find_all("div", class_="columns")
                    )
                print "@@ cols %s"%(cols,)
            self.assertEqual(e, f, "%s != %s (%r)"%(e, f, row_expected))
        return

    def check_list_row_data(self, s, trows_expected):
        trows = s.form.find_all("div", class_="tbody")
        self.assertEqual(len(trows), len(trows_expected))
        for i in range(len(trows_expected)):
            e = trows_expected[i]   # expected
            f = trows[i]            # found
            # <input class="select-box right" name="entity_select" type="checkbox" value="_list/list1"/>
            self.assertEqual(e[0], f.find("input", class_="select-box")["value"])
            for j in range(len(trows_expected[i][1])):
                self.check_row_column(trows[i], j, trows_expected[i][1])
        return

    def check_view_fields(self, s, expect_fields, expect_field_choices):
        trows = s.find_all("div", class_="select-row")
        self.assertEqual(len(trows), len(expect_fields))
        for i in range(len(trows)):
            tcols = trows[i].select("div.columns div.row div.columns")
            self.assertEqual(trows[i].div.input['type'],  "checkbox")
            self.assertEqual(trows[i].div.input['name'],  "View_fields__select_fields")
            self.assertEqual(trows[i].div.input['value'], str(i))
            self.check_select_field(
                tcols[0], "View_fields__%d__Field_id"%i, 
                expect_field_choices, expect_fields[i]
                )
        return

    # Test consistency of type / list / view / field descriptions
    def check_type_list_view(self, type_id, list_id, view_id, type_uri):
        # Read type description - check required fields are present
        type_type = RecordType.load(self.coll1, type_id, altscope="all")
        self.assertEqual(type_type["@type"],                [ANNAL.CURIE.Type])
        self.assertEqual(type_type[ANNAL.CURIE.id],         type_id)
        self.assertEqual(type_type[ANNAL.CURIE.type_id],    "_type")
        self.assertEqual(type_type[ANNAL.CURIE.uri],        type_uri)
        self.assertEqual(type_type[ANNAL.CURIE.type_list],  "_list/"+list_id)
        self.assertEqual(type_type[ANNAL.CURIE.type_view],  "_view/"+view_id)
        # Read type list description
        type_list = RecordList.load(self.coll1, list_id, altscope="all")
        self.assertEqual(type_list["@type"],                    [ANNAL.CURIE.List])
        self.assertEqual(type_list[ANNAL.CURIE.id],             list_id)
        self.assertEqual(type_list[ANNAL.CURIE.type_id],        "_list")
        self.assertEqual(type_list[ANNAL.CURIE.display_type],   "_enum_list_type/List")
        self.assertEqual(type_list[ANNAL.CURIE.default_type],   "_type/"+type_id)
        self.assertEqual(type_list[ANNAL.CURIE.default_view],   "_view/"+view_id)
        self.assertEqual(type_list[ANNAL.CURIE.record_type],    type_uri)
        self.assertIn(ANNAL.CURIE.list_entity_selector,         type_list)
        # Read type view description
        type_view = RecordView.load(self.coll1, view_id, altscope="all")
        self.assertEqual(type_view["@type"],                    [ANNAL.CURIE.View])
        self.assertEqual(type_view[ANNAL.CURIE.id],             view_id)
        self.assertEqual(type_view[ANNAL.CURIE.type_id],        "_view")
        self.assertEqual(type_view[ANNAL.CURIE.record_type],    type_uri)
        self.assertIn(ANNAL.CURIE.open_view,                    type_view)
        # Read and check fields used in list and view displays
        # print "@@ l: " + type_list[ANNAL.CURIE.id]
        self.check_type_fields(type_id, type_uri, type_list[ANNAL.CURIE.list_fields])
        # print "@@ v: " + type_view[ANNAL.CURIE.id]
        self.check_type_fields(type_id, type_uri, type_view[ANNAL.CURIE.view_fields])
        return

    # Test consistency of field descriptions for a given type
    def check_type_fields(self, type_id, type_uri, view_fields):
        # print "@@ t: " + type_id
        for f in view_fields:
            field_id    = extract_entity_id(f[ANNAL.CURIE.field_id])
            # print "@@ f: " + field_id
            view_field  = RecordField.load(self.coll1, field_id, altscope="all")
            render_type = extract_entity_id(view_field[ANNAL.CURIE.field_render_type])
            value_type  = extract_entity_id(view_field[ANNAL.CURIE.field_value_type])
            try:
                self.assertEqual(view_field["@type"], [ANNAL.CURIE.Field])
                self.assertEqual(view_field[ANNAL.CURIE.id],      field_id)
                self.assertEqual(view_field[ANNAL.CURIE.type_id], "_field")
                self.assertIn(ANNAL.CURIE.property_uri,           view_field)
                self.assertIn(ANNAL.CURIE.field_render_type,      view_field)
                self.assertIn(ANNAL.CURIE.field_value_mode,       view_field)
                self.assertIn(ANNAL.CURIE.field_value_type,       view_field)
                self.assertIn(ANNAL.CURIE.field_placement,        view_field)
                self.assertIn(ANNAL.CURIE.placeholder,            view_field)
                self.assertIn(ANNAL.CURIE.field_placement,        view_field)
                if ANNAL.CURIE.field_entity_type in view_field:
                    if view_field[ANNAL.CURIE.field_entity_type] != "":
                        # Field is restricted to named type
                        # @@TODO: allow for subtypes?
                        self.assertEqual(
                            view_field[ANNAL.CURIE.field_entity_type], type_uri
                            )
                if is_repeat_field_render_type(render_type):
                    # Check extra fields
                    self.assertIn(ANNAL.CURIE.field_fields,        view_field)
                    self.assertIn(ANNAL.CURIE.repeat_label_add,    view_field)
                    self.assertIn(ANNAL.CURIE.repeat_label_delete, view_field)
                    self.assertIn(ANNAL.CURIE.field_value_type,    view_field)
                    # Check field list
                    field_list      = view_field[ANNAL.CURIE.field_fields]
                    repeat_type_uri = view_field[ANNAL.CURIE.field_value_type]
                    # print "@@ subfields: " + repeat_type_uri
                    self.check_type_fields("_field", repeat_type_uri, field_list)
                    # print "@@ subfields: end"
                enum_types = (
                    [ "Type", "View", "List", "Field"
                    , "Enum", "Enum_optional"
                    ])
                if render_type in enum_types:
                    self.assertIn(ANNAL.CURIE.field_ref_type, view_field)
            except Exception as e:
                log.warning("check_type_fields error %s, field_id %s, render_type %s"%(e, field_id, render_type))
                raise
        return

    # --------------------------------------------------------------------------
    # Test cases
    # --------------------------------------------------------------------------

    # Site front page
    def test_site_view(self):
        u = site_view_url()
        s = self.get_page(u)
        # Check displayed collections (check site setup)
        self.assertEqual(s.title.string, "Annalist data notebook test site")
        trows = s.select("form > div > div > div")
        self.assertEqual(len(trows), 5)
        for i in (1,2,3):
            tcols = trows[i+1].find_all("div", class_="view-value")
            colln = "coll%d"%i
            self.assertEqual(tcols[0].a.string,  colln)
            self.assertEqual(tcols[0].a['href'], collection_view_url(colln))
        return

    # Collection data view
    def test_get_site_data_view(self):
        collection_url = collection_view_url(coll_id=layout.SITEDATA_ID)
        u = collectiondata_url(coll_id=layout.SITEDATA_ID)
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertEqual(r.context['coll_id'],          layout.SITEDATA_ID)
        self.assertEqual(r.context['type_id'],          layout.COLL_TYPEID)
        self.assertEqual(r.context['entity_id'],        layout.SITEDATA_ID)
        self.assertEqual(r.context['orig_id'],          layout.SITEDATA_ID)
        self.assertEqual(r.context['action'],           "view")
        self.assertEqual(r.context['continuation_url'], "")
        self.assertEqual(
            r.context['entity_data_ref'],      
            collection_url+layout.COLL_META_REF
            )
        self.assertEqual(
            r.context['entity_data_ref_json'], 
            collection_url+layout.COLL_META_REF+"?type=application/json"
            )
        return

    # Collection data content negotiation
    def test_get_site_data_json(self):
        """
        Request collection data as JSON-LD
        """
        collection_url = collection_view_url(coll_id=layout.SITEDATA_ID)
        u = collectiondata_url(coll_id=layout.SITEDATA_ID)
        r = self.client.get(u, HTTP_ACCEPT="application/ld+json")
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        v = r['Location']
        self.assertEqual(v, TestHostUri+collection_url+layout.COLL_META_REF)
        r = self.client.get(v)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        return
 
    # Test collection view
    def test_collection_view(self):
        u = collection_view_url(coll_id="coll1")
        s = self.get_page(u)

        self.assertEqual(s.title.string, "List entities with type information - Collection coll1")
        self.assertEqual(s.h2.string, "List entities with type information")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "_list/Default_list_all")

        thead = s.form.find("div", class_="thead").find("div", class_="row").find_all("div", class_="columns")
        self.assertEqual(thead[0].span.string, "Id")
        self.assertEqual(thead[1].span.string, "Type")
        self.assertEqual(thead[2].span.string, "Label")
        # test_default_list_all performs other relevant tests
        return

    # Test configuration view
    def test_collection_edit(self):
        u = collection_edit_url(coll_id="coll1")
        s = self.get_page(u)
        self.assertEqual(s.h2.string, "Customize collection: Collection coll1")
        local_types_expected = make_field_choices(
            [ ("type1", "RecordType coll1/type1")
            , ("type2", "RecordType coll1/type2")
            ])
        local_lists_expected = make_field_choices(
            [ ("list1", "RecordList coll1/list1")
            , ("list2", "RecordList coll1/list2")
            ])
        local_views_expected = make_field_choices(
            [ ("view1", "RecordView coll1/view1")
            , ("view2", "RecordView coll1/view2")
            ])
        self.check_select_field(s, "typelist", local_types_expected, None)
        self.check_select_field(s, "listlist", local_lists_expected, None)
        self.check_select_field(s, "viewlist", local_views_expected, None)
        return

    # Default list all
    # Same as test_collection_view but using different URI
    def test_default_list_all(self):
        u = collection_entity_list_url(coll_id="coll1", list_id="Default_list_all")
        s = self.get_page(u)

        self.assertEqual(s.h2.string, "List entities with type information")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "_list/Default_list_all")

        thead = s.form.find("div", class_="thead").find("div", class_="row").find_all("div", class_="columns")
        self.assertEqual(thead[0].span.string, "Id")
        self.assertEqual(thead[1].span.string, "Type")
        self.assertEqual(thead[2].span.string, "Label")
        trows_expected = (
            [ [ "_list/list1",    ["list1",    "List",              "RecordList coll1/list1"] ]
            , [ "_list/list2",    ["list2",    "List",              "RecordList coll1/list2"] ]
            , [ "_type/type1",    ["type1",    "Type",              "RecordType coll1/type1"] ]
            , [ "_type/type2",    ["type2",    "Type",              "RecordType coll1/type2"] ]
            , [ "_user/testuser", ["testuser", "User permissions",  "Test User"] ]
            , [ "_view/view1",    ["view1",    "View",              "RecordView coll1/view1"] ]
            , [ "_view/view2",    ["view2",    "View",              "RecordView coll1/view2"] ]
            , [ "type1/entity1",  ["entity1",  "RecordType coll1/type1", "Entity coll1/type1/entity1"] ]
            , [ "type1/entity2",  ["entity2",  "RecordType coll1/type1", "Entity coll1/type1/entity2"] ]
            , [ "type1/entity3",  ["entity3",  "RecordType coll1/type1", "Entity coll1/type1/entity3"] ]
            , [ "type2/entity1",  ["entity1",  "RecordType coll1/type2", "Entity coll1/type2/entity1"] ]
            , [ "type2/entity2",  ["entity2",  "RecordType coll1/type2", "Entity coll1/type2/entity2"] ]
            , [ "type2/entity3",  ["entity3",  "RecordType coll1/type2", "Entity coll1/type2/entity3"] ]
            ])
        self.check_list_row_data(s, trows_expected)
        # @@TODO: check entity and type links
        #         check_list_row_links?  or add 3rd element to each trows_expected
        return

    # Default list
    def test_default_list(self):
        u = collection_entity_list_url(coll_id="coll1", list_id="Default_list")
        s = self.get_page(u)

        self.assertEqual(s.h2.string, "List entities")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "_list/Default_list")

        thead = s.form.find("div", class_="thead").find("div", class_="row").find_all("div", class_="columns")
        self.assertEqual(thead[0].span.string, "Id")
        self.assertEqual(thead[1].span.string, "Label")

        trows_expected = (
            [ [ "_list/list1",    ["list1",    "RecordList coll1/list1"] ]
            , [ "_list/list2",    ["list2",    "RecordList coll1/list2"] ]
            , [ "_type/type1",    ["type1",    "RecordType coll1/type1"] ]
            , [ "_type/type2",    ["type2",    "RecordType coll1/type2"] ]
            , [ "_user/testuser", ["testuser", "Test User"] ]
            , [ "_view/view1",    ["view1",    "RecordView coll1/view1"] ]
            , [ "_view/view2",    ["view2",    "RecordView coll1/view2"] ]
            , [ "type1/entity1",  ["entity1",  "Entity coll1/type1/entity1"] ]
            , [ "type1/entity2",  ["entity2",  "Entity coll1/type1/entity2"] ]
            , [ "type1/entity3",  ["entity3",  "Entity coll1/type1/entity3"] ]
            , [ "type2/entity1",  ["entity1",  "Entity coll1/type2/entity1"] ]
            , [ "type2/entity2",  ["entity2",  "Entity coll1/type2/entity2"] ]
            , [ "type2/entity3",  ["entity3",  "Entity coll1/type2/entity3"] ]
            ])
        self.check_list_row_data(s, trows_expected)
        return

    # --------------------------------------------------------------------------
    # Test site data for types
    # --------------------------------------------------------------------------

    # Test type / list / view / field consistency for RecordType
    def test_recordtype_type_list_view(self):
        self.check_type_list_view("_type", "Type_list", "Type_view", ANNAL.CURIE.Type)
        return

    # List types using type list
    def test_type_list(self):
        u = collection_entity_list_url(
            coll_id="coll1", list_id="Type_list", type_id="_type", scope="all"
            )
        s = self.get_page(u)

        self.assertEqual(s.h2.string, "Entity types")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "_list/Type_list")

        thead = (
            s.form.find("div", class_="thead")
                  .find("div", class_="row")
                  .find_all("div", class_="columns")
            )
        self.assertEqual(thead[0].span.string, "Id")
        self.assertEqual(thead[1].span.string, "Type URI")
        self.assertEqual(thead[2].span.string, "Label")

        trows_expected = (
            [ [ "_type/_coll",             
                [ "_coll",             "annal:Collection",       "Collection"            ] ]
            , [ "_type/_enum_list_type",   
                [ "_enum_list_type",   "annal:Enum_list_type",   "List display type"     ] ]
            , [ "_type/_enum_render_type", 
                [ "_enum_render_type", "annal:Enum_render_type", "Field render type"     ] ]
            , [ "_type/_enum_value_mode",  
                [ "_enum_value_mode",  "annal:Enum_value_mode",  "Field value mode"      ] ]
            , [ "_type/_enum_value_type",  
                [ "_enum_value_type",  "annal:Enum_value_type",  "Field value type"      ] ]
            , [ "_type/_field",            
                [ "_field",            "annal:Field",            "Field"                 ] ]
            , [ "_type/_group",            
                [ "_group",            "annal:Field_group",      "Field group"           ] ]
            , [ "_type/_list",             
                [ "_list",             "annal:List",             "List"                  ] ]
            , [ "_type/_type",             
                [ "_type",             "annal:Type",             "Type"                  ] ]
            , [ "_type/_user",             
                [ "_user",             "annal:User",             "User permissions"      ] ]
            , [ "_type/_view",             
                [ "_view",             "annal:View",             "View"                  ] ]
            , [ "_type/_vocab",            
                [ "_vocab",            "annal:Vocabulary",       "Namespace"             ] ]
            , [ "_type/Default_type",      
                [ "Default_type",      "annal:Default_type",     "Default record"        ] ]
            , [ "_type/type1",             
                [ "type1",             "(@@no value@@)",         "RecordType coll1/type1"] ]
            , [ "_type/type2",             
                [ "type2",             "(@@no value@@)",         "RecordType coll1/type2"] ]
            ])
        self.check_list_row_data(s, trows_expected)
        return

    # Create/edit type using type view
    def test_type_edit_new(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_type",
            view_id="Type_view", action="new"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", None)
        self.check_input_type_value(s, "Type_label", "text", None)
        self.check_input_type_value(s, "Type_comment", "textarea", None)
        self.check_input_type_value(s, "Type_uri", "text", None)
        self.check_select_field(
            s, "Type_view",   
            no_selection("(view id)") + self.views_expected, "_view/Default_view"
            )
        self.check_select_field(
            s, "Type_list",
            no_selection("(list id)") + self.lists_expected, "_list/Default_list"
            )
        self.check_select_field(s, "view_choice", self.views_expected, "_view/Type_view")
        return

    # Edit/view type view
    def test_view_edit_type_view(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_view", entity_id="Type_view",
            view_id="View_view", action="edit"
            )
        s = self.get_page(u)
        expect_field_choices = no_selection("(field sel)") + get_site_type_fields_sorted()
        expect_fields = (
            [ "_field/Type_id"
            , "_field/Type_label"
            , "_field/Type_comment"
            , "_field/Type_uri"
            , "_field/Type_supertype_uris"
            , "_field/Type_view"
            , "_field/Type_list"
            , "_field/Type_aliases"
            ])
        self.check_view_fields(s, expect_fields, expect_field_choices)
        self.check_select_field(s, "view_choice", self.views_expected, "_view/View_view")
        return

    # Edit/view type list
    def test_view_edit_type_list(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_list", entity_id="Type_list",
            view_id="List_view", action="edit"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", "Type_list")
        self.check_input_type_value(s, "List_label", "text", None)
        self.check_input_type_value(s, "List_comment", "textarea", None)
        self.check_input_type_value(s, "List_entity_selector", "text", "'annal:Type' in [@type]")
        self.check_input_type_value(s, "List_target_type", "text", "annal:Type")
        self.check_select_field(
            s, "List_type", 
            self.list_types_expected, 
            "_enum_list_type/List"
            )
        self.check_select_field(
            s, "List_default_type", 
            no_selection("(default entity type)") + self.types_expected, 
            "_type/_type"
            )
        self.check_select_field(
            s, "List_default_view", 
            no_selection("(view id)") + self.views_expected, 
            "_view/Type_view"
            )
        return

    # --------------------------------------------------------------------------
    # Test site data for lists
    # --------------------------------------------------------------------------

    # Test type / list / view / field consistency for RecordList
    def test_recordlist_type_list_view(self):
        self.check_type_list_view("_list", "List_list", "List_view", ANNAL.CURIE.List)
        return

    # List lists using list list
    def test_list_list(self):
        u = collection_entity_list_url(
            coll_id="coll1", list_id="List_list", type_id="_list", scope='all'
            )
        s = self.get_page(u)

        self.assertEqual(s.h2.string, "List definitions")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "_list/List_list")

        thead = s.form.find("div", class_="thead").find("div", class_="row").find_all("div", class_="columns")
        self.assertEqual(thead[0].span.string, "Id")
        self.assertEqual(thead[1].span.string, "Label")

        trows_expected = (
            [ [ "_list/Default_list",       ["Default_list",        "List entities"                      ] ]
            , [ "_list/Default_list_all",   ["Default_list_all",    "List entities with type information"] ]
            , [ "_list/Enum_list_all",      ["Enum_list_all",       "List enumeration values and types"  ] ]
            , [ "_list/Field_group_list",   ["Field_group_list",    "Field groups"                       ] ]
            , [ "_list/Field_list",         ["Field_list",          "Field definitions"                  ] ]
            , [ "_list/List_list",          ["List_list",           "List definitions"                   ] ]
            , [ "_list/Type_list",          ["Type_list",           "Entity types"                       ] ]
            , [ "_list/User_list",          ["User_list",           "User permissions"                   ] ]
            , [ "_list/View_list",          ["View_list",           "View definitions"                   ] ]
            , [ "_list/Vocab_list",         ["Vocab_list",          "Vocabulary namespaces"              ] ]
            , [ "_list/list1",              ["list1",               "RecordList coll1/list1"             ] ]
            , [ "_list/list2",              ["list2",               "RecordList coll1/list2"             ] ]   
            ])
        self.check_list_row_data(s, trows_expected)
        return

    # Create/edit list using list view
    def test_list_edit_new(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_list",
            view_id="List_view", action="new"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", None)
        self.check_input_type_value(s, "List_label", "text", None)
        self.check_input_type_value(s, "List_comment", "textarea", None)
        self.check_select_field(
            s, "List_default_type", 
            no_selection("(default entity type)") + self.types_expected, 
            "_type/Default_type"
            )
        self.check_select_field(
            s, "List_default_view", 
            no_selection("(view id)") + self.views_expected, 
            "_view/Default_view"
            )
        self.check_input_type_value(s, "List_entity_selector", "text", "ALL")
        self.check_input_type_value(s, "List_target_type", "text", "")
        self.check_select_field(s, "view_choice", self.views_expected, "_view/List_view")
        return

    # Edit/view list view
    def test_view_edit_list_view(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_view", entity_id="List_view",
            view_id="View_view", action="edit"
            )
        s = self.get_page(u)
        expect_field_choices = no_selection("(field sel)") + get_site_list_fields_sorted()
        expect_fields = (
            [ "_field/List_id"
            , "_field/List_type"
            , "_field/List_label"
            , "_field/List_comment"
            , "_field/List_default_type"
            , "_field/List_default_view"
            , "_field/List_entity_selector"
            , "_field/List_target_type"
            , "_field/List_fields"
            ])
        self.check_view_fields(s, expect_fields, expect_field_choices)
        self.check_select_field(s, "view_choice", self.views_expected, "_view/View_view")
        return

    # Edit/view list list
    def test_view_edit_list_list(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_list", entity_id="List_list",
            view_id="List_view", action="edit"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", "List_list")
        self.check_input_type_value(s, "List_label", "text", None)
        self.check_input_type_value(s, "List_comment", "textarea", None)
        self.check_input_type_value(s, "List_entity_selector", "text", "'annal:List' in [@type]")
        self.check_input_type_value(s, "List_target_type", "text", "annal:List")
        self.check_select_field(s, "List_type", self.list_types_expected, "_enum_list_type/List")
        self.check_select_field(
            s, "List_default_type", 
            no_selection("(default entity type)") + self.types_expected, 
            "_type/_list"
            )
        self.check_select_field(
            s, "List_default_view", 
            no_selection("(view id)") + self.views_expected, 
            "_view/List_view"
            )
        return

    # --------------------------------------------------------------------------
    # Test site data for views
    # --------------------------------------------------------------------------

    # Test type / list / view / field consistency for RecordView
    def test_recordlist_type_view_view(self):
        self.check_type_list_view("_view", "View_list", "View_view", ANNAL.CURIE.View)
        return

    # List views using view list
    def test_view_list(self):
        u = collection_entity_list_url(
            coll_id="coll1", list_id="View_list", type_id="_view", scope='all'
            )
        s = self.get_page(u)

        self.assertEqual(s.h2.string, "View definitions")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "_list/View_list")

        thead = s.form.find("div", class_="thead").find("div", class_="row").find_all("div", class_="columns")
        self.assertEqual(thead[0].span.string, "Id")
        self.assertEqual(thead[1].span.string, "Label")

        trows_expected = (
            [ [ "_view/Collection_view",    ["Collection_view",     "Collection metadata"   ] ]
            , [ "_view/Default_view",       ["Default_view",        "Default record view"   ] ]
            , [ "_view/Enum_view",          ["Enum_view",           "Enumerated value view" ] ]
            , [ "_view/Field_group_view",   ["Field_group_view",    "Field group definition"] ]
            , [ "_view/Field_view",         ["Field_view",          "Field definition"      ] ]
            , [ "_view/List_view",          ["List_view",           "List definition"       ] ]
            , [ "_view/Type_view",          ["Type_view",           "Type definition"       ] ]
            , [ "_view/User_view",          ["User_view",           "User permissions"      ] ]
            , [ "_view/View_view",          ["View_view",           "View definition"       ] ]
            , [ "_view/Vocab_view",         ["Vocab_view",          "Vocabulary namespace"  ] ]
            , [ "_view/view1",              ["view1",               "RecordView coll1/view1"] ]
            , [ "_view/view2",              ["view2",               "RecordView coll1/view2"] ]   
            ])
        self.check_list_row_data(s, trows_expected)
        return

    # Create/edit view using view view
    def test_view_edit_new(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_view",
            view_id="View_view", action="new"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", None)
        self.check_input_type_value(s, "View_label", "text", None)
        self.check_input_type_value(s, "View_comment", "textarea", None)
        self.check_input_type_value(s, "View_target_type", "text", None)
        self.check_input_type_value(s, "View_edit_view", "checkbox", "Yes")
        self.check_select_field(s, "view_choice", self.views_expected, "_view/View_view")
        return

    # Edit/view view view
    def test_view_edit_view_view(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_view", entity_id="View_view",
            view_id="View_view", action="edit"
            )
        s = self.get_page(u)
        expect_field_choices = no_selection("(field sel)") + get_site_view_fields_sorted()
        expect_fields        = (
            [ "_field/View_id"
            , "_field/View_label"
            , "_field/View_comment"
            , "_field/View_target_type"
            , "_field/View_edit_view"
            , "_field/View_fields"
            ])
        self.check_view_fields(s, expect_fields, expect_field_choices)
        self.check_select_field(s, "view_choice", self.views_expected, "_view/View_view")
        return

    # Edit/view view list
    def test_view_edit_view_list(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_list", entity_id="View_list",
            view_id="List_view", action="edit"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", "View_list")
        self.check_input_type_value(s, "List_label", "text", None)
        self.check_input_type_value(s, "List_comment", "textarea", None)
        self.check_select_field(s, "List_type", self.list_types_expected, "_enum_list_type/List")
        self.check_select_field(
            s, "List_default_type", 
            no_selection("(default entity type)") + self.types_expected, 
            "_type/_view"
            )
        self.check_select_field(
            s, "List_default_view", 
            no_selection("(view id)") + self.views_expected, 
            "_view/View_view"
            )
        self.check_input_type_value(
            s, "List_entity_selector", "text", "'annal:View' in [@type]"
            )
        self.check_input_type_value(
            s, "List_target_type", "text", "annal:View"
            )
        return

    # --------------------------------------------------------------------------
    # Test site data for field groups
    # --------------------------------------------------------------------------

    # Test type / list / view / field consistency for RecordGroup
    def test_recordlist_type_group_view(self):
        self.check_type_list_view(
            "_group", "Field_group_list", "Field_group_view", ANNAL.CURIE.Field_group
            )
        return

    # List groups using group list
    def test_group_list(self):
        u = collection_entity_list_url(
            coll_id="coll1", list_id="Field_group_list", type_id="_group", scope="all"
            )
        s = self.get_page(u)

        self.assertEqual(s.h2.string, "Field groups")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "_list/Field_group_list")

        thead = s.form.find("div", class_="thead").find("div", class_="row").find_all("div", class_="columns")
        self.assertEqual(thead[0].span.string, "Id")
        self.assertEqual(thead[1].span.string, "Label")

        trows_expected = (
            [
            ])
        self.check_list_row_data(s, trows_expected)
        return

    # Create/edit group using group view
    # NOTE: this presents a form for editing the group entity
    def test_group_edit_new(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_group",
            view_id="Field_group_view", action="new"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", None)
        self.check_input_type_value(s, "Group_label", "text", None)
        self.check_input_type_value(s, "Group_comment", "textarea", None)
        self.check_input_type_value(s, "Group_target_type", "text", None)
        expect_field_choices = no_selection("(field sel)") + get_site_group_fields_sorted()
        expect_fields = []
        self.check_view_fields(s, expect_fields, expect_field_choices)
        self.check_select_field(
            s, "view_choice", self.views_expected, "_view/Field_group_view"
            )
        return

    # Edit/view group view
    # NOTE: this presents a form for editing the group view entity
    def test_view_edit_group_view(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_view", entity_id="Field_group_view",
            view_id="View_view", action="edit"
            )
        s = self.get_page(u)
        expect_field_choices = no_selection("(field sel)") + get_site_group_fields_sorted()
        expect_fields = (
            [ "_field/Group_id"
            , "_field/Group_label"
            , "_field/Group_comment"
            , "_field/Group_target_type"
            , "_field/Group_fields"
            ])
        self.check_view_fields(s, expect_fields, expect_field_choices)
        self.check_select_field(s, "view_choice", self.views_expected, "_view/View_view")
        return

    # Edit/view group list
    def test_view_edit_group_list(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_list", entity_id="Field_group_list",
            view_id="List_view", action="edit"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", "Field_group_list")
        self.check_input_type_value(s, "List_label", "text", None)
        self.check_input_type_value(s, "List_comment", "textarea", None)
        self.check_input_type_value(
            s, "List_entity_selector", "text", "'annal:Field_group' in [@type]"
            )
        self.check_input_type_value(s, "List_target_type", "text", "annal:Field_group")
        self.check_select_field(s, "List_type", self.list_types_expected, "_enum_list_type/List")
        self.check_select_field(
            s, "List_default_type", 
            no_selection("(default entity type)") + self.types_expected, 
            "_type/_group"
            )
        self.check_select_field(
            s, "List_default_view", 
            no_selection("(view id)") + self.views_expected, 
            "_view/Field_group_view"
            )
        return

    # --------------------------------------------------------------------------
    # Test site data for record fields
    # --------------------------------------------------------------------------

    # Test type / list / view / field consistency for RecordField
    def test_recordlist_type_field_view(self):
        self.check_type_list_view("_field", "Field_list", "Field_view", ANNAL.CURIE.Field)
        return

    # List fields using field list
    def test_field_list(self):
        u = collection_entity_list_url(
            coll_id="coll1", list_id="Field_list", type_id="_field", scope="all"
            )
        s = self.get_page(u)

        self.assertEqual(s.h2.string, "Field definitions")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "_list/Field_list")

        thead = (
            s.form.find("div", class_="thead")
            .find("div", class_="row")
            .find_all("div", class_="columns")
            )
        self.assertEqual(thead[0].span.string, "Id")
        self.assertEqual(thead[1].span.string, "Render type")
        self.assertEqual(thead[2].span.string, "Value type")
        self.assertEqual(thead[3].span.string, "Label")

        trows_expected = (
            #     Field selector                      Field id             Render type      Value type   Field label (?)
            [ [ "_field/Coll_comment",              ["Coll_comment",      "Markdown rich text", 
                                                                                            "annal:Richtext"         ] ]
            , [ "_field/Coll_default_list_id",      ["Coll_default_list_id", 
                                                                          "Display text",   "annal:Text"             ] ]
            , [ "_field/Coll_default_view_entity",  ["Coll_default_view_entity", 
                                                                          "Display text",   "annal:Text"             ] ]
            , [ "_field/Coll_default_view_id",      ["Coll_default_view_id", 
                                                                          "Display text",   "annal:Text"             ] ]
            , [ "_field/Coll_default_view_type",    ["Coll_default_view_type", 
                                                                          "Display text",   "annal:Text"             ] ]
            , [ "_field/Coll_parent",               ["Coll_parent",       "Optional entity reference",  
                                                                                            "annal:EntityRef"        ] ]
            , [ "_field/Coll_software_version",     ["Coll_software_version", 
                                                                          "Display text",   "annal:Text"             ] ]
            , [ "_field/Entity_comment",            ["Entity_comment",    "Markdown rich text", "annal:Richtext"     ] ]
            , [ "_field/Entity_id",                 ["Entity_id",         "Entity Id",      "annal:EntityRef"        ] ]
            , [ "_field/Entity_label",              ["Entity_label",      "Short text",     "annal:Text"             ] ]
            , [ "_field/Entity_see_also",           ["Entity_see_also",   "Web link",       "rdfs:Resource"          ] ]
            , [ "_field/Entity_see_also_r",         ["Entity_see_also_r", "Field group set as table", 
                                                                                        "annal:Entity_see_also_list" ] ]
            , [ "_field/Entity_type",               ["Entity_type",       "Entity type Id", "annal:EntityRef"        ] ]
            , [ "_field/Entity_uri",                ["Entity_uri",        "Identifier",     "annal:Identifier"       ] ]
            , [ "_field/Enum_uri",                  ["Enum_uri",          "Identifier",     "annal:Identifier"       ] ]
            , [ "_field/Field_default",             ["Field_default",     "Short text",     "annal:Text"             ] ]
            , [ "_field/Field_entity_type",         ["Field_entity_type", "Identifier",     "annal:Identifier"       ] ]
            , [ "_field/Field_fieldref",            ["Field_fieldref",    "Identifier",     "annal:Identifier"       ] ]
            , [ "_field/Field_fields",              ["Field_fields",      "Field group sequence as table",
                                                                                            "annal:Field_list"       ] ]
            , [ "_field/Field_groupref",            ["Field_groupref",    "Optional/new entity reference", 
                                                                                            "annal:EntityRef"        ] ]
            , [ "_field/Field_help",                ["Field_help",        "Markdown rich text",  "annal:Richtext"    ] ]
            , [ "_field/Field_id",                  ["Field_id",          "Entity Id",      "annal:EntityRef"        ] ]
            , [ "_field/Field_label",               ["Field_label",       "Short text",     "annal:Text"             ] ]
            , [ "_field/Field_missing",             ["Field_missing",     "Short text",     "annal:Text"             ] ]
            , [ "_field/Field_placeholder",         ["Field_placeholder", "Short text",     "annal:Text"             ] ]
            , [ "_field/Field_placement",           ["Field_placement",   "Position/size",  "annal:Placement"        ] ]
            , [ "_field/Field_property",            ["Field_property",    "Identifier",     "annal:Identifier"       ] ]
            , [ "_field/Field_render_type",         ["Field_render_type", "Required entity reference",  
                                                                                            "annal:EntityRef"        ] ]
            , [ "_field/Field_repeat_label_add",    ["Field_repeat_label_add", "Short text", "annal:Text"            ] ]
            , [ "_field/Field_repeat_label_delete", ["Field_repeat_label_delete", "Short text", "annal:Text"         ] ]
            , [ "_field/Field_restrict",            ["Field_restrict",    "Short text",     "annal:Text"             ] ]
            , [ "_field/Field_subfield_placement",  ["Field_subfield_placement",
                                                                          "Position/size",  "annal:Placement"        ] ]
            , [ "_field/Field_subfield_property",   ["Field_subfield_property",
                                                                          "Identifier",     "annal:Identifier"       ] ]
            , [ "_field/Field_subfield_sel",        ["Field_subfield_sel",
                                                                          "Optional/new entity reference", 
                                                                                            "annal:Field"            ] ]
            , [ "_field/Field_tooltip",             ["Field_tooltip",     "Multiline text", "annal:Longtext"         ] ]
            , [ "_field/Field_typeref",             ["Field_typeref",     "Optional/new entity reference", 
                                                                                            "annal:EntityRef"        ] ]
            , [ "_field/Field_value_mode",          ["Field_value_mode",  "Required entity reference",  
                                                                                            "annal:EntityRef"        ] ]
            , [ "_field/Field_value_type",          ["Field_value_type",  "Identifier",     "annal:Identifier"       ] ]
            , [ "_field/Group_comment",             ["Group_comment"             ] ]
            , [ "_field/Group_field_placement",     ["Group_field_placement"     ] ]
            , [ "_field/Group_field_property",      ["Group_field_property"      ] ]
            , [ "_field/Group_field_sel",           ["Group_field_sel"           ] ]
            , [ "_field/Group_fields",              ["Group_fields"              ] ]
            , [ "_field/Group_id",                  ["Group_id"                  ] ]
            , [ "_field/Group_label",               ["Group_label"               ] ]
            , [ "_field/Group_target_type",         ["Group_target_type"         ] ]
            , [ "_field/List_choice",               ["List_choice"               ] ]
            , [ "_field/List_comment",              ["List_comment"              ] ]
            , [ "_field/List_default_type",         ["List_default_type"         ] ]
            , [ "_field/List_default_view",         ["List_default_view"         ] ]
            , [ "_field/List_entity_selector",      ["List_entity_selector"      ] ]
            , [ "_field/List_field_placement",      ["List_field_placement"      ] ]
            , [ "_field/List_field_property",       ["List_field_property"       ] ]
            , [ "_field/List_field_sel",            ["List_field_sel"            ] ]
            , [ "_field/List_fields",               ["List_fields"               ] ]
            , [ "_field/List_id",                   ["List_id"                   ] ]
            , [ "_field/List_label",                ["List_label"                ] ]
            , [ "_field/List_target_type",          ["List_target_type"          ] ]
            , [ "_field/List_type",                 ["List_type"                 ] ]
            , [ "_field/Type_alias_source",         ["Type_alias_source"         ] ]
            , [ "_field/Type_alias_target",         ["Type_alias_target"         ] ]
            , [ "_field/Type_aliases",              ["Type_aliases"              ] ]
            , [ "_field/Type_comment",              ["Type_comment"              ] ]
            , [ "_field/Type_id",                   ["Type_id"                   ] ]
            , [ "_field/Type_label",                ["Type_label"                ] ]
            , [ "_field/Type_list",                 ["Type_list"                 ] ]
            , [ "_field/Type_supertype_uri",        ["Type_supertype_uri"        ] ]
            , [ "_field/Type_supertype_uris",       ["Type_supertype_uris"       ] ]
            , [ "_field/Type_uri",                  ["Type_uri"                  ] ]
            , [ "_field/Type_view",                 ["Type_view"                 ] ]
            , [ "_field/User_description",          ["User_description"          ] ]
            , [ "_field/User_id",                   ["User_id"                   ] ]
            , [ "_field/User_name",                 ["User_name"                 ] ]
            , [ "_field/User_permissions",          ["User_permissions"          ] ]
            , [ "_field/User_uri",                  ["User_uri"                  ] ]
            , [ "_field/View_choice",               ["View_choice"               ] ]
            , [ "_field/View_comment",              ["View_comment"              ] ]
            , [ "_field/View_edit_view",            ["View_edit_view"            ] ]
            , [ "_field/View_field_placement",      ["View_field_placement"      ] ]
            , [ "_field/View_field_property",       ["View_field_property"       ] ]
            , [ "_field/View_field_sel",            ["View_field_sel"            ] ]
            , [ "_field/View_fields",               ["View_fields"               ] ]
            , [ "_field/View_id",                   ["View_id"                   ] ]
            , [ "_field/View_label",                ["View_label"                ] ]
            , [ "_field/View_target_type",          ["View_target_type"          ] ]
            , [ "_field/Vocab_id",                  ["Vocab_id"                  ] ]
            , [ "_field/Vocab_uri",                 ["Vocab_uri"                 ] ]
            ])
        self.check_list_row_data(s, trows_expected)
        return

    # Create/edit field using field view
    def test_field_edit_new(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_field",
            view_id="Field_view", action="new"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id",         "text", None)
        self.check_input_type_value(s, "Field_value_type",  "text", "")
        self.check_input_type_value(s, "Field_label",       "text", "")
        self.check_input_type_value(s, "Field_help",        "textarea", None)
        self.check_input_type_value(s, "Field_placeholder", "text", "")
        self.check_input_type_value(s, "Field_tooltip",     "textarea", None)
        self.check_input_type_value(s, "Field_property",    "text", "")
        # self.check_input_type_value(s, "Field_placement", "text", "")
        self.check_select_field(
            s, "Field_placement", 
            [FieldChoice("",label="(field position and size)")]+self.placements_expected, 
            ""
            )
        self.check_input_type_value(s, "Field_default",     "text", None)
        self.check_input_type_value(s, "Field_entity_type", "text", "")
        self.check_input_type_value(s, "Field_restrict",    "text", "")
        self.check_select_field(
            s, "Field_render_type",   self.render_types_expected, "_enum_render_type/Text"
            )
        self.check_select_field(
            s, "Field_typeref", 
            [FieldChoice("",label="(no type selected)")]+self.types_expected, 
            ""
            )
        self.check_select_field(
            s, "view_choice", self.views_expected, "_view/Field_view"
            )
        return

    # Edit/view field view
    def test_view_edit_field_view(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_view", entity_id="Field_view",
            view_id="View_view", action="edit"
            )
        s = self.get_page(u)
        expect_field_choices = no_selection("(field sel)") + get_site_field_fields_sorted()
        expect_fields = (
            [ "_field/Field_id"
            , "_field/Field_render_type"
            , "_field/Field_label"
            , "_field/Field_help"
            , "_field/Field_property"
            , "_field/Field_placement"
            , "_field/Field_value_type"
            , "_field/Field_value_mode"
            , "_field/Field_entity_type"
            , "_field/Field_typeref"
            , "_field/Field_fieldref"
            , "_field/Field_default"
            , "_field/Field_placeholder"
            , "_field/Field_tooltip"
            , "_field/Field_fields"
            , "_field/Field_repeat_label_add"
            , "_field/Field_repeat_label_delete"
            , "_field/Field_restrict"
            ])
        self.check_view_fields(s, expect_fields, expect_field_choices)
        self.check_select_field(s, "view_choice", self.views_expected, "_view/View_view")
        return

    # Edit/view field list
    def test_view_edit_field_list(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_list", entity_id="Field_list",
            view_id="List_view", action="edit"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", "Field_list")
        self.check_input_type_value(s, "List_label", "text", None)
        self.check_input_type_value(s, "List_comment", "textarea", None)
        self.check_input_type_value(s, "List_target_type", "text", "annal:Field")
        self.check_input_type_value(
            s, "List_entity_selector", "text", "'annal:Field' in [@type]"
            )
        self.check_select_field(
            s, "List_default_type", 
            no_selection("(default entity type)") + self.types_expected, 
            "_type/_field"
            )
        self.check_select_field(
            s, "List_default_view", 
            no_selection("(view id)") + self.views_expected, 
            "_view/Field_view"
            )
        return

    # --------------------------------------------------------------------------
    # Test site data for vocabulary namespaces
    # --------------------------------------------------------------------------

    # Test type / list / view / field consistency for RecordGroup
    def test_recordlist_type_group_view(self):
        self.check_type_list_view(
            "_vocab", "Vocab_list", "Vocab_view", ANNAL.CURIE.Vocabulary
            )
        return

    # List vocabularies using vocab list
    def test_vocab_list(self):
        u = collection_entity_list_url(
            coll_id="coll1", list_id="Vocab_list", type_id="_vocab", scope="all"
            )
        s = self.get_page(u)
        self.assertEqual(s.h2.string, "Vocabulary namespaces")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "_list/Vocab_list")
        thead = s.form.find("div", class_="thead").find("div", class_="row").find_all("div", class_="columns")
        self.assertEqual(thead[0].span.string, "Id")
        self.assertEqual(thead[1].span.string, "Label")
        trows_expected = (
            [ [ "_vocab/annal",                    ["annal",    "Vocabulary namespace for Annalist-defined terms"] ]
            , [ "_vocab/owl",                      ["owl",      "OWL ontology namespace"] ]
            , [ "_vocab/rdf",                      ["rdf",      "RDF core namespace"] ]
            , [ "_vocab/rdfs",                     ["rdfs",     "RDF schema namespace"] ]
            , [ "_vocab/xsd",                      ["xsd",      "XML Schema datatypes namespace"] ]
            ])
        self.check_list_row_data(s, trows_expected)
        return

    # Create/edit Vocabulary using vocab view
    def test_vocab_edit_new(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_vocab",
            view_id="Vocab_view", action="new"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id",              "text",     None)
        self.check_input_type_value(s, "Entity_label",           "text",     None)
        self.check_input_type_value(s, "Entity_comment",         "textarea", None)
        self.check_input_type_value(s, "Vocab_uri",              "text",     None)
        self.check_select_field(
            s, "view_choice", self.views_expected, "_view/Vocab_view"
            )
        return

    # Edit/view vocabulary view
    def test_view_edit_vocab_view(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_view", entity_id="Vocab_view",
            view_id="View_view", action="edit"
            )
        s = self.get_page(u)
        expect_field_choices = no_selection("(field sel)") + get_site_vocab_fields_sorted()
        expect_fields = (
            [ "_field/Vocab_id"
            , "_field/Entity_label"
            , "_field/Entity_comment"
            , "_field/Vocab_uri"
            , "_field/Entity_see_also_r"
            ])
        self.check_view_fields(s, expect_fields, expect_field_choices)
        self.check_select_field(s, "view_choice", self.views_expected, "_view/View_view")
        return

    # Edit/view vocabulary list
    def test_view_edit_vocab_list(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_list", entity_id="Vocab_list",
            view_id="List_view", action="edit"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", "Vocab_list")
        self.check_input_type_value(s, "List_label", "text", None)
        self.check_input_type_value(s, "List_comment", "textarea", None)
        self.check_input_type_value(
            s, "List_entity_selector", "text", "'annal:Vocabulary' in [@type]"
            )
        self.check_input_type_value(s, "List_target_type", "text", "annal:Vocabulary")
        self.check_select_field(s, "List_type", self.list_types_expected, "_enum_list_type/List")
        self.check_select_field(
            s, "List_default_type", 
            no_selection("(default entity type)") + self.types_expected, 
            "_type/_vocab"
            )
        self.check_select_field(
            s, "List_default_view", 
            no_selection("(view id)") + self.views_expected, 
            "_view/Vocab_view"
            )
        return

    # --------------------------------------------------------------------------
    # Test site data for user permissions
    # --------------------------------------------------------------------------

    # Test type / list / view / field consistency for user
    def test_recorduser_type_field_view(self):
        self.check_type_list_view("_user", "User_list", "User_view", ANNAL.CURIE.User)
        return

    # List user permissions using user list
    def test_user_list(self):
        u = collection_entity_list_url(
            coll_id="coll1", list_id="User_list", type_id="_user", scope="all"
            )
        s = self.get_page(u)

        self.assertEqual(s.h2.string, "User permissions")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "_list/User_list")

        thead = s.form.find("div", class_="thead").find("div", class_="row").find_all("div", class_="columns")
        self.assertEqual(thead[0].span.string, "User Id")
        self.assertEqual(thead[1].span.string, "User URI")
        self.assertEqual(thead[2].span.string, "Permissions")

        trows_expected = (
            [ [ "_user/_default_user_perms",  
                [ "_default_user_perms",  "annal:User/_default_user_perms"
                , "VIEW"]
              ]
            , [ "_user/_unknown_user_perms",
                [ "_unknown_user_perms",  "annal:User/_unknown_user_perms"
                , "VIEW"
                ] 
              ]
            # , [ "_user/admin", 
            #     [ "admin", "mailto:admin@localhost"
            #     , "VIEW CREATE UPDATE DELETE CONFIG CREATE_COLLECTION DELETE_COLLECTION ADMIN"
            #     ]
            #   ]
            , [ "_user/testuser",
                [ "testuser", "mailto:testuser@test.example.com"
                , "VIEW CREATE UPDATE DELETE CONFIG ADMIN"
                ]
              ]
            ])
        # print s.prettify()
        self.check_list_row_data(s, trows_expected)
        return

    # Create/edit user using user view
    def test_user_edit_new(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_user",
            view_id="User_view", action="new"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", None)
        self.check_input_type_value(s, "User_name", "text", None)
        self.check_input_type_value(s, "User_description", "textarea", None)
        self.check_input_type_value(s, "User_uri", "text", None)
        self.check_input_type_value(s, "User_permissions", "text", None)
        self.check_select_field(s, "view_choice", self.views_expected, "_view/User_view")
        return

    # Edit/view user view
    def test_view_edit_user_view(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_view", entity_id="User_view",
            view_id="View_view", action="edit"
            )
        s = self.get_page(u)
        expect_field_choices = no_selection("(field sel)") + get_site_user_fields_sorted()
        expect_fields = (
            [ "_field/User_id"
            , "_field/User_name"
            , "_field/User_description"
            , "_field/User_uri"
            , "_field/User_permissions"
            ])
        self.check_view_fields(s, expect_fields, expect_field_choices)
        self.check_select_field(s, "view_choice", self.views_expected, "_view/View_view")
        return

    # Edit/view user list
    def test_view_edit_user_list(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_list", entity_id="User_list",
            view_id="List_view", action="edit"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", "User_list")
        self.check_input_type_value(s, "List_label", "text", None)
        self.check_input_type_value(s, "List_comment", "textarea", None)
        self.check_input_type_value(s, "List_target_type", "text", "annal:User")
        self.check_input_type_value(
            s, "List_entity_selector", "text", "'annal:User' in [@type]"
            )
        self.check_select_field(
            s, "List_default_type", 
            no_selection("(default entity type)") + self.types_expected, 
            "_type/_user"
            )
        self.check_select_field(
            s, "List_default_view", 
            no_selection("(view id)") + self.views_expected, 
            "_view/User_view"
            )
        return

# End.



"""
Tests for generic EntityData non-editing view
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest

import logging
log = logging.getLogger(__name__)

from django.conf                    import settings
from django.db                      import models
from django.http                    import QueryDict
from django.contrib.auth.models     import User
from django.test                    import TestCase # cf. https://docs.djangoproject.com/en/dev/topics/testing/tools/#assertions
from django.test.client             import Client

from utils.SuppressLoggingContext   import SuppressLogging

from annalist.identifiers           import RDF, RDFS, ANNAL
from annalist                       import layout

from annalist.models.entitytypeinfo import EntityTypeInfo
from annalist.models.site           import Site
from annalist.models.collection     import Collection
from annalist.models.recordtype     import RecordType
from annalist.models.recordlist     import RecordList
from annalist.models.recordview     import RecordView
from annalist.models.recordfield    import RecordField
from annalist.models.recordtypedata import RecordTypeData
from annalist.models.entitydata     import EntityData

from annalist.views.entityedit              import GenericEntityEditView
from annalist.views.form_utils.fieldchoice  import FieldChoice

from AnnalistTestCase       import AnnalistTestCase
from tests                  import TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
from init_tests             import init_annalist_test_site, init_annalist_test_coll, resetSitedata
from entity_testutils       import (
    collection_create_values,
    site_dir, collection_dir, 
    continuation_url_param,
    collection_view_url,
    collection_edit_url,
    collection_entity_view_url,
    site_title,
    render_select_options,
    render_choice_options,
    create_test_user,
    context_view_field,
    check_default_view_context_fields,
    check_type_view_context_fields
    )
from entity_testtypedata    import (
    recordtype_dir, 
    recordtype_edit_url,
    recordtype_create_values, 
    )
from entity_testentitydata  import (
    recorddata_dir,  entitydata_dir,
    entity_url, entitydata_edit_url, 
    entitydata_list_type_url,
    entitydata_value_keys, entitydata_create_values, entitydata_values,
    entitydata_delete_confirm_form_data,
    entitydata_default_view_form_data,
    entitydata_recordtype_view_context_data, entitydata_recordtype_view_form_data,
    default_fields, default_label, default_comment, error_label,
    layout_classes
    )
from entity_testsitedata    import (
    get_site_types, get_site_types_sorted, get_site_types_linked,
    get_site_lists, get_site_lists_sorted, get_site_lists_linked,
    get_site_views, get_site_views_sorted, get_site_views_linked,
    get_site_list_types, get_site_list_types_sorted,
    get_site_field_groups, get_site_field_groups_sorted, 
    get_site_fields, get_site_fields_sorted, 
    get_site_field_types, get_site_field_types_sorted, 
    )
from entity_testviewdata    import recordview_url
from entity_testlistdata    import recordlist_url

#   -----------------------------------------------------------------------------
#
#   GenericEntityViewView tests
#
#   -----------------------------------------------------------------------------

class GenericEntityViewViewTest(AnnalistTestCase):

    def setUp(self):
        init_annalist_test_site()
        init_annalist_test_coll()
        self.testsite = Site(TestBaseUri, TestBaseDir)
        self.testcoll = Collection.create(self.testsite, "testcoll", collection_create_values("testcoll"))
        self.testtype = RecordType.create(self.testcoll, "testtype", recordtype_create_values("testtype"))
        self.testdata = RecordTypeData.create(self.testcoll, "testtype", {})
        self.no_options = [ FieldChoice('', label="(no options)") ]
        self.no_view_id = [ FieldChoice('', label="(view id)") ]
        self.no_list_id = [ FieldChoice('', label="(list id)") ]
        self.view_options = get_site_views_linked("testcoll")
        self.list_options = get_site_lists_linked("testcoll")
        # Login and permissions
        create_test_user(self.testcoll, "testuser", "testpassword")
        self.client = Client(HTTP_HOST=TestHost)
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        return

    def tearDown(self):
        # resetSitedata()
        return

    @classmethod
    def tearDownClass(cls):
        resetSitedata()
        return

    #   -----------------------------------------------------------------------------
    #   Helpers
    #   -----------------------------------------------------------------------------

    def _create_entity_data(self, entity_id, update="Entity"):
        "Helper function creates entity data with supplied entity_id"
        e = EntityData.create(self.testdata, entity_id, 
            entitydata_create_values(entity_id, update=update)
            )
        return e    

    def _check_entity_data_values(self, entity_id, type_id="testtype", update="Entity", update_dict=None):
        "Helper function checks content of form-updated record type entry with supplied entity_id"
        # log.info("_check_entity_data_values: type_id %s, entity_id %s"%(type_id, entity_id))
        typeinfo = EntityTypeInfo(self.testcoll, type_id)
        self.assertTrue(typeinfo.entityclass.exists(typeinfo.entityparent, entity_id))
        e = typeinfo.entityclass.load(typeinfo.entityparent, entity_id)
        self.assertEqual(e.get_id(), entity_id)
        self.assertEqual(e.get_view_url(""), TestHostUri + entity_url("testcoll", type_id, entity_id))
        v = entitydata_values(entity_id, type_id=type_id, update=update)
        if update_dict:
            v.update(update_dict)
            for k in update_dict:
                if update_dict[k] is None:
                    v.pop(k, None)
        # log.info(e.get_values())
        self.assertDictionaryMatch(e.get_values(), v)
        return e

    #   -----------------------------------------------------------------------------
    #   Form rendering and response tests
    #   -----------------------------------------------------------------------------

    def test_get_default_form_no_login(self):
        self.client.logout()
        u = entity_url("testcoll", "testtype", "entity1")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        return

    def test_post_default_form_use_view(self):
        self._create_entity_data("entityuseview")
        self.assertTrue(EntityData.exists(self.testdata, "entityuseview"))
        f = entitydata_default_view_form_data(
                entity_id="entityuseview", action="view",
                use_view="_view/Type_view", 
                )
        f.pop('entity_id', None)
        u = entitydata_edit_url(
            "view", "testcoll", "testtype", "entityuseview", view_id="Default_view"
            )
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        self.assertEqual(r.content,       "")
        v = TestHostUri + entitydata_edit_url(
            "view", "testcoll", "testtype", entity_id="entityuseview", view_id="Type_view"
            )
        c = continuation_url_param("/testsite/c/testcoll/d/testtype/")
        self.assertIn(v, r['location'])
        self.assertIn(c, r['location'])
        self._check_entity_data_values("entityuseview")
        return

    def test_post_default_form_default_view(self):
        # Set default entity view, then ensure collection view redirects to it
        self._create_entity_data("entitydefaultview")
        f = entitydata_default_view_form_data(
                entity_id="entitydefaultview", action="view",
                default_view="default_view", 
                )
        f.pop('entity_id', None)
        u = entitydata_edit_url(
            action="view", view_id="Default_view",
            coll_id="testcoll", type_id="testtype", entity_id="entitydefaultview"
            )
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        self.assertEqual(r.content,       "")
        v = TestHostUri + u
        self.assertIn(v, r['location'])
        ih = "info_head=Action%20completed"
        im = (
            "info_message="+
            "Default%20view%20for%20collection%20testcoll%20changed%20to%20"+
            "Default_view/testtype/entitydefaultview"
            )
        self.assertIn(ih, r['location'])
        self.assertIn(im, r['location'])
        # Get collection root and check redirect to entity view
        u2 = collection_view_url(coll_id="testcoll")
        r2 = self.client.get(u2)
        self.assertEqual(r2.status_code,   302)
        self.assertEqual(r2.reason_phrase, "FOUND")
        self.assertEqual(r2.content,       "")
        v2 = TestHostUri + entitydata_edit_url(
            coll_id="testcoll", view_id="Default_view", type_id="testtype", entity_id="entitydefaultview"
            )
        self.assertEqual(v2, r2['location'])
        return

    def test_post_default_form_use_view_no_login(self):
        self.client.logout()
        f = entitydata_default_view_form_data(
                entity_id="entityuseview", action="view",
                use_view="_view/Type_view", 
                )
        u = entity_url("testcoll", "testtype", "entity1")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        return

    def test_get_view_form_no_login(self):
        self.client.logout()
        u = entitydata_edit_url("view", "testcoll", "testtype", entity_id="entity1", view_id="Type_view")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        return

    def test_get_view_form_rendering(self):
        u = entitydata_edit_url("view", "testcoll", "testtype", entity_id="entity1", view_id="Type_view")
        r = self.client.get(u+"?continuation_url=/xyzzy/")
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # log.info(r.content)     #@@
        entity_label = """Entity testcoll/testtype/entity1"""
        view_label   = """Type definition"""
        permalink    = """<a href="./" class="permalink">&#x1F517;</a>"""
        entity_title = "%s - %s - Collection testcoll"%(entity_label, view_label)
        self.assertContains(r, "<title>%s</title>"%(entity_title,))
        self.assertContains(r, '<h2 class="page-heading">%s %s</h2>'%(view_label, permalink), html=True)
        cont_uri = "?continuation_url=%s"%u + "%3Fcontinuation_url=/xyzzy/"
        field_vals = default_fields(
            coll_id="testcoll", type_id="testtype", entity_id="00000001",
            entity_base      = "/testsite/c/testcoll/d/testtype/entity1",
            entity_url       = "/testsite/c/testcoll/d/testtype/entity1/" + cont_uri,
            default_view_url = "/testsite/c/testcoll/d/_view/Default_view/" + cont_uri,
            default_list_url = "/testsite/c/testcoll/d/_list/Default_list/" + cont_uri,
            tooltip1="", # 'title="%s"'%r.context['fields'][0]['field_help'],
            tooltip2="", # 'title="%s"'%r.context['fields'][1]['field_help'],
            tooltip3="", # 'title="%s"'%r.context['fields'][2]['field_help'],
            tooltip4="", # 'title="%s"'%r.context['fields'][3]['field_help'],
            tooltip5="", # 'title="%s"'%r.context['fields'][4]['field_help'],
            tooltip6="", # 'title="%s"'%r.context['fields'][5]['field_help'],
            tooltip7="", # 'title="%s"'%r.context['fields'][6]['field_help'],
            button_subtype_tip=
                "Create a subtype of the current type.  "+
                "(View and list type identifiers are copied from the current type; "+
                "the URI of the current type is inserted as a supertype URI of the new type; "+
                "other fields are taken from the corresponding '_initial_values' record, "+
                "and may be extended or modified later.)"+
                "",
            )
        formrow1 = """
            <div class="small-12 medium-6 columns" %(tooltip1)s>
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Type Id</span>
                </div>
                <div class="%(input_classes)s">
                  <a href="%(entity_url)s">entity1</a>
                </div>
              </div>
            </div>
            """%field_vals(width=6)
        formrow2 = """
            <div class="small-12 columns" %(tooltip2)s>
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Label</span>
                </div>
                <div class="%(input_classes)s">
                  <span>Entity testcoll/testtype/entity1</span>
                </div>
              </div>
            </div>
            """%field_vals(width=12)
        formrow3 = """
            <div class="small-12 columns" %(tooltip3)s>
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Comment</span>
                </div>
                <div class="%(input_classes)s">
                  <span class="markdown">
                    <p>Entity coll testcoll, type testtype, entity entity1</p>
                  </span>
                </div>
              </div>
            </div>
            """%field_vals(width=12)
        formrow4 = """
            <div class="small-12 columns" %(tooltip4)s>
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Type URI</span>
                </div>
                <div class="%(input_classes)s">
                  <span>&nbsp;</span>
                </div>
              </div>
            </div>
            """%field_vals(width=12)
        formrow5 = """
            <div class="small-12 columns" %(tooltip5)s>
              <div class="grouprow row">
                <div class="%(group_label_classes)s">
                  <span>Supertype URIs</span>
                </div>
                <div class="%(group_placeholder_classes)s">
                  <span>(None specified)</span>
                </div>
              </div>
            </div>
            """%field_vals(width=12)
        formrow6 = """
            <div class="small-12 medium-6 columns" %(tooltip6)s>
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Default view</span>
                </div>
                <div class="%(input_classes)s">
                  <a href="%(default_view_url)s">Default record view</a>
                </div>
              </div>
            </div>
            """%field_vals(width=6)
        formrow7 = """
            <div class="small-12 medium-6 columns" %(tooltip7)s>
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Default list</span>
                </div>
                <div class="%(input_classes)s">
                  <a href="%(default_list_url)s">List entities</a>
                </div>
              </div>
            </div>
            """%field_vals(width=6)
        formrow8a = """
            <div class="%(space_classes)s">
              <div class="row">
                <div class="small-12 columns">
                  &nbsp;
                </div>
              </div>
            </div>
            """%field_vals(width=2)
        formrow8b = """
            <div class="%(button_wide_classes)s">
              <div class="row">
                <div class="%(button_left_classes)s">
                  <input type="submit" name="edit"  value="Edit"
                         title="Edit entity data." />
                  <input type="submit" name="copy"  value="Copy"
                         title="Copy, then edit entity data as new entity." />
                  <input type="submit" name="close" value="Close"
                         title="Return to previous page." />
                </div>
              </div>
            </div>
            """%field_vals(width=4)
        formrow8c = """
            <div class="%(button_wide_classes)s">
              <div class="row">
                <div class="%(button_r_med_up_classes)s">
                  <!-- <input type="submit" name="open_view"    value="View description" /> -->
                  <input type="submit" name="Define_subtype" value="Define subtype"
                         title="%(button_subtype_tip)s" />
                  <input type="submit" name="default_view" value="Set default view"
                         title="Select this display as the default view for collection 'testcoll'." />
                  <input type="submit" name="customize"    value="Customize"
                         title="Open 'Customize' view for collection 'testcoll'." />
                </div>
              </div>
            </div>
            """%field_vals(width=6)
        formrow9a = ("""
            <div class="row view-value-row">
              <div class="%(label_classes)s">
                <span>Choose view</span>
              </div>
              <div class="%(input_classes)s">
                <div class="row">
                  <div class="small-9 columns">
                  """+
                    render_choice_options(
                      "view_choice",
                      self.view_options,
                      "_view/Type_view")+
                  """
                  </div>
                  <div class="small-3 columns">
                    <input type="submit" name="use_view"      value="Show view" />
                  </div>
                </div>
              </div>
            </div>
            """)%field_vals(width=6)
        formrow9b = ("""
            <div class="row view-value-row">
              <div class="link-bar small-12 columns">
                <a href="%(entity_base)s/entity_data.ttl" title="Retrieve underlying data as Turtle">
                  Turtle
                </a>
                <a href="%(entity_base)s/entity_data.jsonld" title="Retrieve underlying data as JSON-LD">
                  JSON-LD
                </a>
                <a href="%(entity_base)s/entity_data.jsonld?type=application/json" title="Display underlying JSON data">
                  <img src="/static/images/get_the_data_88x31.png" alt="get_the_data">
                </a>
              </div>
            </div>
            """)%field_vals(width=6)
        formrow10 = """
            <div class="%(button_right_classes)s">
              <div class="row">
                <div class="small-12 columns">
                  <input type="submit" name="new_type"      value="New type" />
                  <input type="submit" name="new_view"      value="New view" />
                  <input type="submit" name="new_field"     value="New field type" />
                  <input type="submit" name="new_group"     value="New field group" />
                </div>
              </div>
            </div>
            """%field_vals(width=6)
        # log.info(r.content)     #@@
        self.assertContains(r, formrow1,  html=True)
        self.assertContains(r, formrow2,  html=True)
        self.assertContains(r, formrow3,  html=True)
        self.assertContains(r, formrow4,  html=True)
        self.assertContains(r, formrow5,  html=True)
        self.assertContains(r, formrow6,  html=True)
        self.assertContains(r, formrow7,  html=True)
        self.assertContains(r, formrow8a, html=True)
        self.assertContains(r, formrow8b, html=True)
        self.assertContains(r, formrow8c, html=True)
        self.assertContains(r, formrow9a, html=True)
        self.assertContains(r, formrow9b, html=True)
        # New buttons hidden (for now)
        # self.assertContains(r, formrow10, html=True)
        return

    def test_get_view(self):
        # Note - this test uses Type_view to display en entity of type "testtype"
        u = entitydata_edit_url("edit", "testcoll", "testtype", entity_id="entity1", view_id="Type_view")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        check_type_view_context_fields(self, r, 
            action="edit",
            entity_id="entity1", orig_entity_id="entity1",
            type_id="testtype",
            type_label="Entity testcoll/testtype/entity1",
            type_comment="Entity coll testcoll, type testtype, entity entity1",
            type_uri="",
            type_supertype_uris=[],
            type_view="Default_view", type_view_options=self.no_view_id + self.view_options,
            type_list="Default_list", type_list_options=self.no_list_id + self.list_options,
            type_aliases=[],
            )
        return

    # Entity data view
    def test_get_data_view(self):
        u = entitydata_edit_url("edit", "testcoll", "testtype", entity_id="entity1")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        check_default_view_context_fields(self, r, 
            action="edit",
            type_id="testtype",
            entity_id="entity1", orig_entity_id="entity1",
            entity_label="Entity testcoll/testtype/entity1",
            entity_comment="Entity coll testcoll, type testtype, entity entity1",
            entity_data_ref=entity_url("testcoll", "testtype", entity_id="entity1")+layout.ENTITY_DATA_FILE,
            view_id="Default_view"
            )
        u = collection_entity_view_url(coll_id="testcoll", type_id="testtype", entity_id="entity1")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertEqual(r.context['coll_id'],          "testcoll")
        self.assertEqual(r.context['type_id'],          "testtype")
        self.assertEqual(r.context['entity_id'],        "entity1")
        self.assertEqual(r.context['orig_id'],          "entity1")
        self.assertEqual(r.context['action'],           "view")
        self.assertEqual(r.context['continuation_url'], "")
        self.assertEqual(
            r.context['entity_data_ref'],      
            u+layout.ENTITY_DATA_FILE
            )
        self.assertEqual(
            r.context['entity_data_ref_json'], 
            u+layout.ENTITY_DATA_FILE+"?type=application/json"
            )
        return

    # Entity data content negotiation
    def test_get_data_json(self):
        """
        Request collection data as JSON-LD
        """
        u = collection_entity_view_url(coll_id="testcoll", type_id="testtype", entity_id="entity1")
        r = self.client.get(u, HTTP_ACCEPT="application/ld+json")
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        v = r['Location']
        self.assertEqual(v, TestHostUri+u+layout.ENTITY_DATA_FILE)
        return

    def test_get_view_no_collection(self):
        u = entitydata_edit_url("view", "no_collection", "_field", entity_id="entity1", view_id="Type_view")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   404)
        self.assertEqual(r.reason_phrase, "Not found")
        self.assertContains(r, "Collection no_collection does not exist", status_code=404)
        return

    def test_get_view_no_type(self):
        u = entitydata_edit_url("edit", "testcoll", "no_type", entity_id="entity1", view_id="Type_view")
        with SuppressLogging(logging.WARNING):
            r = self.client.get(u)
        self.assertEqual(r.status_code,   404)
        self.assertEqual(r.reason_phrase, "Not found")
        self.assertContains(r, "Record type no_type in collection testcoll does not exist", status_code=404)
        return

    def test_get_view_no_view(self):
        u = entitydata_edit_url("edit", "testcoll", "_field", entity_id="entity1", view_id="no_view")
        with SuppressLogging(logging.WARNING):
            r = self.client.get(u)
        self.assertEqual(r.status_code,   404)
        self.assertEqual(r.reason_phrase, "Not found")
        self.assertContains(r, "Record view no_view in collection testcoll does not exist", status_code=404)
        return

    def test_get_view_no_entity(self):
        u = entitydata_edit_url("view", "testcoll", "testtype", entity_id="entitynone", view_id="Type_view")
        r = self.client.get(u+"?continuation_url=/xyzzy/")
        self.assertEqual(r.status_code,   404)
        self.assertEqual(r.reason_phrase, "Not found")
        self.assertContains(r, "<title>Annalist error</title>", status_code=404)
        self.assertContains(r, "<h3>404: Not found</h3>", status_code=404)
        # log.debug(r.content)
        err_label = error_label("testcoll", "testtype", "entitynone")
        self.assertContains(r, "<p>Entity %s does not exist</p>"%err_label, status_code=404)
        return

    #   -----------------------------------------------------------------------------
    #   Form response tests
    #   -----------------------------------------------------------------------------

    def test_post_view_entity_close(self):
        self._create_entity_data("entityview")
        f = entitydata_default_view_form_data(entity_id="entityview", action="view", close="Close")
        u = entitydata_edit_url("view", "testcoll", "testtype", entity_id="entityview", view_id="Default_view")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        self.assertEqual(r.content,       "")
        self.assertEqual(r['location'], TestHostUri + entitydata_list_type_url("testcoll", "testtype"))
        self._check_entity_data_values("entityview")
        return

    def test_post_view_entity_edit(self):
        self._create_entity_data("entityview")
        f = entitydata_default_view_form_data(entity_id="entityview", action="view", edit="Edit")
        u = entitydata_edit_url("view", "testcoll", "testtype", entity_id="entityview", view_id="Default_view")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        self.assertEqual(r.content,       "")
        e = TestHostUri + entitydata_edit_url(
            "edit", "testcoll", "testtype", entity_id="entityview", view_id="Default_view"
            )
        l = continuation_url_param(entitydata_list_type_url("testcoll", "testtype"))
        c = continuation_url_param(u, prev_cont=l)
        self.assertIn(e, r['location'])
        self.assertIn(c, r['location'])
        # 'http://test.example.com/testsite/c/testcoll/v/Default_view/testtype/entityview/!edit
        #   ?continuation_url=/testsite/c/testcoll/v/Default_view/testtype/entityview/!view
        #   %3Fcontinuation_url=/testsite/c/testcoll/d/testtype/'
        return

    def test_post_view_entity_copy(self):
        self._create_entity_data("entityview")
        f = entitydata_default_view_form_data(entity_id="entityview", action="view", copy="Copy")
        u = entitydata_edit_url("view", "testcoll", "testtype", entity_id="entityview", view_id="Default_view")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        self.assertEqual(r.content,       "")
        e = TestHostUri + entitydata_edit_url(
            "copy", "testcoll", "testtype", entity_id="entityview", view_id="Default_view"
            )
        l = continuation_url_param(entitydata_list_type_url("testcoll", "testtype"))
        c = continuation_url_param(u, prev_cont=l)
        self.assertIn(e, r['location'])
        self.assertIn(c, r['location'])
        return

# End.



"""
Entity resource access (imported and uploaded files).

This module contains functions to help access resource data that is 
attached to an entity (or collection).

It contains logic for accessing raw JSON-LD data, or accessing that data converted to 
some other format (e.g. Turtle).
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2017, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import sys
import os
import json
import StringIO
import logging
log = logging.getLogger(__name__)

from rdflib                             import Graph, URIRef, Literal

from annalist                           import message
from annalist                           import layout

from annalist.models.entitytypeinfo     import EntityTypeInfo

# Resource info data for built-in entity data

collection_fixed_json_resources = (
    [ { "resource_name": layout.COLL_META_FILE,     "resource_dir": layout.COLL_BASE_DIR, 
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.COLL_PROV_FILE,     "resource_dir": layout.COLL_BASE_DIR, 
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.COLL_CONTEXT_FILE,  "resource_dir": layout.COLL_BASE_DIR, 
                                                    "resource_type": "application/ld+json" }
    ])

entity_fixed_json_resources = (
    [ { "resource_name": layout.COLL_META_FILE,     "resource_dir": layout.COLL_BASE_DIR, 
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.COLL_PROV_FILE,     "resource_dir": layout.COLL_BASE_DIR, 
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.TYPE_META_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.TYPE_PROV_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.LIST_META_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.LIST_PROV_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.VIEW_META_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.VIEW_PROV_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.GROUP_META_FILE,    "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.GROUP_PROV_FILE,    "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.FIELD_META_FILE,    "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.FIELD_PROV_FILE,    "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.VOCAB_META_FILE,    "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.VOCAB_PROV_FILE,    "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.USER_META_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.USER_PROV_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.ENUM_META_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.ENUM_PROV_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.TYPEDATA_META_FILE, "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.ENTITY_DATA_FILE,   "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.ENTITY_PROV_FILE,   "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    ])

entity_list_json_resources = (
    # [ { "resource_name": layout.COLL_META_FILE,     "resource_dir": layout.COLL_BASE_DIR, 
    #                                                 "resource_type": "application/ld+json" }
    # , { "resource_name": layout.COLL_PROV_FILE,     "resource_dir": layout.COLL_BASE_DIR, 
    #                                                 "resource_type": "application/ld+json" }
    # , { "resource_name": layout.COLL_CONTEXT_FILE,  "resource_dir": layout.COLL_BASE_DIR, 
    #                                                 "resource_type": "application/ld+json" }
    [ { "resource_name": layout.ENTITY_LIST_FILE,   "resource_dir": ".", 
                                                    "resource_type": "application/ld+json" }
    ])

# Resource access functions

def entity_resource_file(entity, resource_info):
    """
    Return a file object that reads out the content of a resource attached to a specified entity.
    """
    return entity.resource_file(resource_info["resource_path"])

def json_resource_file(baseurl, jsondata, resource_info):
    """
    Return a file object that reads out a JSON version of the supplied entity values data. 
    """
    response_file = StringIO.StringIO()
    json.dump(jsondata, response_file, indent=2, separators=(',', ': '), sort_keys=True)
    response_file.seek(0)
    return response_file

def turtle_resource_file(baseurl, jsondata, resource_info):
    """
    Return a file object that reads out a Turtle version of the supplied entity values data. 

    baseurl     base URL for resolving relative URI references for Turtle output.
    jsondata    is the data to be formatted and returned.
    links       is an optional array of link values to be added to the HTTP response
                (see method add_link_header for description).
    """
    jsondata_file = json_resource_file(baseurl, jsondata, resource_info)
    g = Graph()
    g = g.parse(source=jsondata_file, publicID=baseurl, format="json-ld")
    response_file = StringIO.StringIO()
    g.serialize(destination=response_file, format='turtle', indent=4)
    response_file.seek(0)
    return response_file

def make_turtle_resource_info(json_resource):
    """
    Return Turtle resource description for fixed JSON resource
    """
    turtle_resource = (
        { "resource_name":      json_resource["resource_name"][0:-7]+".ttl"
        , "resource_dir":       json_resource["resource_dir"]
        , "resource_type":      "text/turtle"
        , "resource_access":    turtle_resource_file
        })
    return turtle_resource

def find_fixed_resource(fixed_json_resources, resource_ref):
    """
    Return a description for the indicated fixed resource from a supplied table,
    or None
    """
    # log.debug("CollectionResourceAccess.find_resource %s/d/%s"%(coll.get_id(), resource_ref))
    for fj in fixed_json_resources:
        if fj["resource_name"] == resource_ref:
            fr = dict(fj, resource_path=os.path.join(fj["resource_dir"]+"/", resource_ref))
            return fr
        ft = make_turtle_resource_info(fj)
        if ft["resource_name"] == resource_ref:
            fr = dict(ft, resource_path=os.path.join(ft["resource_dir"]+"/", resource_ref))
            return fr
    log.debug("EntityResourceAccess.find_fixed_resource: %s not found"%(resource_ref))
    return None

def find_entity_resource(entity, resource_ref):
    """
    Return a description for the indicated entity resource, or None
    """
    log.debug(
        "EntityResourceAccess.find_entity_resource %s/%s/%s"%
        (entity.get_type_id(), entity.get_id(), resource_ref)
        )
    fr = find_fixed_resource(entity_fixed_json_resources, resource_ref)
    if fr:
        return fr
    for t, f in entity.enum_fields():
        # log.debug("find_resource: t %s, f %r"%(t,f))
        if isinstance(f, dict):
            if f.get("resource_name", None) == resource_ref:
                f = dict(f, resource_path=resource_ref)
                return f
    return None

def find_list_resource(type_id, list_id, list_ref):
    """
    Return a description for the indicated entity resource, or None
    """
    log.debug(
        "EntityResourceAccess.find_list_resource %s/%s/%s"%
        (list_id, type_id, list_ref)
        )
    return find_fixed_resource(entity_list_json_resources, list_ref)

# End.
